

from torch import Tensor, nn, functional


# this would be more elegant, but sadly using a stack with multiple sequential models 
# (as opposed to some layers, then a sequential model) 
# seems to silently break loading weights

# class CustomPreprocessing(tf.keras.Sequential):
#     def call(self, x, training):
#         # I add the step manually to the top-level model, but this inner model won't have that same var - could add if needed
#         x = super().call(x, training=True)  # always use training=True
#         tf.summary.image('after_preprocessing_layers', x, step=0)
#         return x


class PermaDropout(nn.modules.dropout._DropoutNd):
    # https://pytorch.org/docs/stable/_modules/torch/nn/modules/dropout.html#Dropout
    def forward(self, input: Tensor) -> Tensor:
            return nn.functional.dropout(input, self.p, True, self.inplace)  # simply replaced self.training with True

# # class PermaRandomTranslation(layers.experimental.preprocessing.RandomTranslation):
# #     def call(self, x, training=None):
# #         return super().call(x, training=True)

# class PermaRandomRotation(tf.keras.layers.experimental.preprocessing.RandomRotation):
#     def call(self, x, training=None):
#         return super().call(x, training=True)

# class PermaRandomFlip(tf.keras.layers.experimental.preprocessing.RandomFlip):
#     def call(self, x, training=None):
#         return super().call(x, training=True)

# class PermaRandomCrop(tf.keras.layers.experimental.preprocessing.RandomCrop):
#     def call(self, x, training=None):
#         return super().call(x, training=True)



import tensorflow as tf
from tensorflow.keras import layers


# this would be more elegant, but sadly using a stack with multiple sequential models 
# (as opposed to some layers, then a sequential model) 
# seems to silently break loading weights

# class CustomPreprocessing(tf.keras.Sequential):
#     def call(self, x, training):
#         # I add the step manually to the top-level model, but this inner model won't have that same var - could add if needed
#         x = super().call(x, training=True)  # always use training=True
#         tf.summary.image('after_preprocessing_layers', x, step=0)
#         return x


class PermaDropout(layers.Dropout):
    def call(self, x, training=None):
        return super().call(x, training=True)  # ME, force dropout on at test time

# class PermaRandomTranslation(layers.experimental.preprocessing.RandomTranslation):
#     def call(self, x, training=None):
#         return super().call(x, training=True)

class PermaRandomRotation(tf.keras.layers.RandomRotation):
    def call(self, x, training=None):
        return super().call(x, training=True)

class PermaRandomFlip(tf.keras.layers.RandomFlip):
    def call(self, x, training=None):
        return super().call(x, training=True)

class PermaRandomCrop(tf.keras.layers.RandomCrop):
    def call(self, x, training=None):
        return super().call(x, training=True)



import logging
import pytorch_lightning as pl

from timm import create_model


class ZoobotEncoder(pl.LightningModule):

    def __init__(self, encoder):
        super().__init__()
        logging.info('ZoobotEncoder: using provided in-memory encoder')
        self.encoder = encoder  # plain pytorch module e.g. Sequential


    def forward(self, x):
        if isinstance(x, list) and len(x) == 1:
            return self(x[0])
        return self.encoder(x)
    
    @classmethod
    def load_from_name(cls, name: str):
        """
        e.g. ZoobotEncoder.load_from_name('hf_hub:mwalmsley/zoobot-encoder-convnext_nano')
        Args:
            name (str): huggingface hub name to load

        Returns:
            nn.Module: timm model
        """
        timm_model = create_model(name, pretrained=True)
        return cls(timm_model)
    


import tensorflow as tf
import tensorflow_probability as tfp



def get_multiquestion_loss(question_index_groups, sum_over_questions=True, reduction=tf.keras.losses.Reduction.SUM):
    """
    Get subclass of tf.keras.losses.Loss which wraps ``calculate_multiquestion_loss`` and sums over batch.

    tf.keras.losses.Reduction.SUM_OVER_BATCH_SIZE will assume the question dimension (10) is also part of the batch, and divide by an extra factor of 10
    tf.keras.losses.Reduction.SUM_OVER_BATCH_SIZE also does not work in distributed training (as batch size is split between replicas)
    tf.keras.losses.Reduction.SUM will simply add everything up, so divide by the global batch size externally with tf.reduce_sum

    Args:
        question_index_groups (list): Answer indices for each question i.e. [(question.start_index, question.end_index), ...] for all questions. Useful for slicing model predictions by question. See :ref:`schemas`.

    Returns:
        MultiquestionLoss: see above.
    """

    class MultiquestionLoss(tf.keras.losses.Loss):

        def call(self, labels, predictions):
            return calculate_multiquestion_loss(labels, predictions, question_index_groups, sum_over_questions)

    return MultiquestionLoss(reduction=reduction) 


def calculate_multiquestion_loss(labels, predictions, question_index_groups, sum_over_questions=True):
    """
    The full decision tree loss used for training GZ DECaLS models

    Negative log likelihood of observing ``labels`` (volunteer answers to all questions)
    from Dirichlet-Multinomial distributions for each question, using concentrations ``predictions``.
    
    Args:
        labels (tf.Tensor): (galaxy, k successes) where k successes dimension is indexed by question_index_groups.
        predictions (tf.Tensor):  Dirichlet concentrations, matching shape of labels
        question_index_groups (list): Answer indices for each question i.e. [(question.start_index, question.end_index), ...] for all questions. Useful for slicing model predictions by question. See :ref:`schemas`.
    
    Returns:
        tf.Tensor: neg. log likelihood of shape (batch, question).
    """
    # very important that question_index_groups is fixed and discrete, else tf.function autograph will mess up 
    q_losses = []
    # will give shape errors if model output dim is not labels dim, which can happen if losses.py substrings are missing an answer
    for q_n in range(len(question_index_groups)):
        q_indices = question_index_groups[q_n]
        q_start = q_indices[0]
        q_end = q_indices[1]

        q_loss = dirichlet_loss(labels[:, q_start:q_end+1], predictions[:, q_start:q_end+1])

        q_losses.append(q_loss)
    
    total_loss_with_question_dim = tf.stack(q_losses, axis=1)

    if sum_over_questions:
        total_loss = tf.reduce_sum(total_loss_with_question_dim, axis=1)  # sum (prob-multiply) across questions
        return total_loss
    else:
        return total_loss_with_question_dim
        # has shape (batch_size) to be aggregated outside this func (take a global batch mean)
        # for distributed context, be very careful about device batch size vs. global batch size and consequences for summing/averaging



def dirichlet_loss(labels_for_q, concentrations_for_q):
    """
    Negative log likelihood of ``labels_for_q`` being drawn from Dirichlet-Multinomial distribution with ``concentrations_for_q`` concentrations.
    This loss is for one question. Sum over multiple questions if needed (assumes independent).
    Applied by :meth:`calculate_multiquestion_loss`, above.

    Args:
        labels_for_q (tf.constant): observed labels (count of volunteer responses) of shape (batch, answer)
        concentrations_for_q (tf.constant): concentrations for multinomial-dirichlet predicting the observed labels, of shape (batch, answer)

    Returns:
        tf.constant: negative log. prob per galaxy, of shape (batch_dim).
    """
    # if you get a dimension mismatch here like [16, 2] vs. [64, 2], for example, check --shard-img-size is correct in train_model.py.
    # images may be being reshaped to the wrong expected size e.g. if they are saved as 32x32, but you put --shard-img-size=64,
    # you will get image batches of shape [N/4, 64, 64, 1] and hence have the wrong number of images vs. labels (and meaningless images)
    # so check --shard-img-size carefully!
    total_count = tf.reduce_sum(labels_for_q, axis=1)

    return get_dirichlet_neg_log_prob(labels_for_q, total_count, concentrations_for_q)


    # # return tf.where(
    # #     tf.math.equal(total_count, 0),  # if total_count is 0 (i.e. no votes)
    # #     tf.constant(0.),  # prob is 1 and (neg) log prob is 0, with no gradients attached. broadcasts
    # #     get_dirichlet_neg_log_prob(labels_for_q, total_count, concentrations_for_q)  # else, calculate neg log prob
    # # )
    # # # slightly inefficient as get_dirichlet_neg_log_prob forward pass done for all, but avoids gradients
    # # # https://www.tensorflow.org/api_docs/python/tf/where
    # # works great, but about 50% slower than optimal

    # # indices_with_zero_counts = tf.where(tf.math.equal(total_count, 0))
    # indices_with_nonzero_counts = tf.where(tf.math.logical_not(tf.math.equal(total_count, 0)))
    
    # # may potentially need to deal with the situation where there are 0 valid indices?

    # total_counts_with_nonzero_counts = tf.gather(total_count, indices_with_nonzero_counts)
    # labels_with_nonzero_counts = tf.gather(labels_for_q, indices_with_nonzero_counts)
    # concentrations_with_nonzero_counts = tf.gather(concentrations_for_q, indices_with_nonzero_counts)
    # neg_log_prob_of_indices_with_nonzero_counts = get_dirichlet_neg_log_prob(labels_with_nonzero_counts, total_counts_with_nonzero_counts, concentrations_with_nonzero_counts)

    # # not needed, scatter_nd has 0 where no value is placed
    # # neg_log_prob_of_indices_with_zero_counts = tf.zeros_like(indices_with_zero_counts)

    # # now mix back together
    # # indices and updates are both shape (<=N, 1)
    # # so output shape must also be (N, 1) i.e. rank 1
    # output_shape = tf.convert_to_tensor(value=(tf.size(total_count), 1), dtype=tf.int64)  # tf.size okay here as size=len for total_counts
    # return tf.scatter_nd(indices_with_nonzero_counts, neg_log_prob_of_indices_with_nonzero_counts, shape=output_shape)



def get_dirichlet_neg_log_prob(labels_for_q, total_count, concentrations_for_q):
    """_summary_

    Args:
        labels_for_q (_type_): _description_
        total_count (_type_): _description_
        concentrations_for_q (_type_): _description_

    Returns:
        _type_: _description_
    """
    dist = tfp.distributions.DirichletMultinomial(total_count, concentrations_for_q, validate_args=True)
    return -dist.log_prob(labels_for_q)  # important minus sign




from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Conv2D, MaxPooling2D
from tensorflow.keras.layers import Activation, Dropout, Flatten, Dense

from zoobot.tensorflow.estimators import define_model

# courtesy https://blog.keras.io/building-powerful-image-classification-models-using-very-little-data.html

def small_cnn(input_shape=(224, 224, 1), output_dim=1):
    # model = Sequential()
    model = define_model.CustomSequential()
    model.add(Conv2D(32, (3, 3), input_shape=input_shape))
    model.add(Activation('relu'))
    model.add(MaxPooling2D(pool_size=(2, 2)))

    model.add(Conv2D(32, (3, 3)))
    model.add(Activation('relu'))
    model.add(MaxPooling2D(pool_size=(2, 2)))

    model.add(Conv2D(64, (3, 3)))
    model.add(Activation('relu'))
    model.add(MaxPooling2D(pool_size=(2, 2)))

    model.add(Flatten())  # this converts our 3D feature maps to 1D feature vectors
    model.add(Dense(64, name='dense64'))
    model.add(Activation('relu'))
    # model.add(Dropout(0.5))
    model.add(Dense(output_dim, name='dense_output'))
    model.add(Activation('sigmoid'))

    return model


# Copied almost exactly from https://github.com/qubvel/efficientnet for convenience, all credit to those authors. 
# Dropout switched to permanantly on via #97 and #364
# Please see the license on that repository.
# - Mike

import collections
import glob
import logging
import math
import os
import string
import time
from functools import partial

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from keras_applications.imagenet_utils import _obtain_input_shape

# use my custom layers to have dropout always on
from zoobot.tensorflow.estimators import custom_layers

IMAGENET_WEIGHTS_PATH ='https://github.com/Callidior/keras-applications/releases/download/efficientnet/'

IMAGENET_WEIGHTS_HASHES = {
    'efficientnet-b0': ('163292582f1c6eaca8e7dc7b51b01c61'
                        '5b0dbc0039699b4dcd0b975cc21533dc',
                        'c1421ad80a9fc67c2cc4000f666aa507'
                        '89ce39eedb4e06d531b0c593890ccff3'),
    'efficientnet-b1': ('d0a71ddf51ef7a0ca425bab32b7fa7f1'
                        '6043ee598ecee73fc674d9560c8f09b0',
                        '75de265d03ac52fa74f2f510455ba64f'
                        '9c7c5fd96dc923cd4bfefa3d680c4b68'),
    'efficientnet-b2': ('bb5451507a6418a574534aa76a91b106'
                        'f6b605f3b5dde0b21055694319853086',
                        '433b60584fafba1ea3de07443b74cfd3'
                        '2ce004a012020b07ef69e22ba8669333'),
    'efficientnet-b3': ('03f1fba367f070bd2545f081cfa7f3e7'
                        '6f5e1aa3b6f4db700f00552901e75ab9',
                        'c5d42eb6cfae8567b418ad3845cfd63a'
                        'a48b87f1bd5df8658a49375a9f3135c7'),
    'efficientnet-b4': ('98852de93f74d9833c8640474b2c698d'
                        'b45ec60690c75b3bacb1845e907bf94f',
                        '7942c1407ff1feb34113995864970cd4'
                        'd9d91ea64877e8d9c38b6c1e0767c411'),
    'efficientnet-b5': ('30172f1d45f9b8a41352d4219bf930ee'
                        '3339025fd26ab314a817ba8918fefc7d',
                        '9d197bc2bfe29165c10a2af8c2ebc675'
                        '07f5d70456f09e584c71b822941b1952'),
    'efficientnet-b6': ('f5270466747753485a082092ac9939ca'
                        'a546eb3f09edca6d6fff842cad938720',
                        '1d0923bb038f2f8060faaf0a0449db4b'
                        '96549a881747b7c7678724ac79f427ed'),
    'efficientnet-b7': ('876a41319980638fa597acbbf956a82d'
                        '10819531ff2dcb1a52277f10c7aefa1a',
                        '60b56ff3a8daccc8d96edfd40b204c11'
                        '3e51748da657afd58034d54d3cec2bac')
}

backend = tf.keras.backend
layers = tf.keras.layers
models = tf.keras.models
utils = tf.keras.utils
keras_utils = tf.keras.utils 

BlockArgs = collections.namedtuple('BlockArgs', [
    'kernel_size', 'num_repeat', 'input_filters', 'output_filters',
    'expand_ratio', 'id_skip', 'strides', 'se_ratio'
])
# defaults will be a public argument for namedtuple in Python 3.7
# https://docs.python.org/3/library/collections.html#collections.namedtuple
BlockArgs.__new__.__defaults__ = (None,) * len(BlockArgs._fields)

DEFAULT_BLOCKS_ARGS = [
    BlockArgs(kernel_size=3, num_repeat=1, input_filters=32, output_filters=16,
              expand_ratio=1, id_skip=True, strides=[1, 1], se_ratio=0.25),
    BlockArgs(kernel_size=3, num_repeat=2, input_filters=16, output_filters=24,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=5, num_repeat=2, input_filters=24, output_filters=40,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=3, num_repeat=3, input_filters=40, output_filters=80,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=5, num_repeat=3, input_filters=80, output_filters=112,
              expand_ratio=6, id_skip=True, strides=[1, 1], se_ratio=0.25),
    BlockArgs(kernel_size=5, num_repeat=4, input_filters=112, output_filters=192,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=3, num_repeat=1, input_filters=192, output_filters=320,
              expand_ratio=6, id_skip=True, strides=[1, 1], se_ratio=0.25)
]

CONV_KERNEL_INITIALIZER = {
    'class_name': 'VarianceScaling',
    'config': {
        'scale': 2.0,
        'mode': 'fan_out',
        # EfficientNet actually uses an untruncated normal distribution for
        # initializing conv layers, but keras.initializers.VarianceScaling use
        # a truncated distribution.
        # We decided against a custom initializer for better serializability.
        'distribution': 'normal'
    }
}

DENSE_KERNEL_INITIALIZER = {
    'class_name': 'VarianceScaling',
    'config': {
        'scale': 1. / 3.,
        'mode': 'fan_out',
        'distribution': 'uniform'
    }
}

def get_swish():

    def swish(x):
        """Swish activation function: x * sigmoid(x).
        Reference: [Searching for Activation Functions](https://arxiv.org/abs/1710.05941)
        """

        return tf.nn.swish(x)

    return swish

def get_dropout():
    """Wrapper over custom dropout. Fix problem of ``None`` shape for tf.keras.
    It is not possible to define FixedDropout class as global object,
    because we do not have modules for inheritance at first time.
    Issue:
        https://github.com/tensorflow/tensorflow/issues/30946
    """
    class FixedDropout(layers.Dropout):
    # class FixedDropout(custom_layers.PermaDropout):  # inherit from mine instead of layers.Dropout
        def _get_noise_shape(self, inputs):
            if self.noise_shape is None:
                return self.noise_shape

            symbolic_shape = tf.shape(inputs)
            noise_shape = [symbolic_shape[axis] if shape is None else shape
                           for axis, shape in enumerate(self.noise_shape)]
            return tuple(noise_shape)



    return FixedDropout


def round_filters(filters, width_coefficient, depth_divisor):
    """Round number of filters based on width multiplier."""

    filters *= width_coefficient
    new_filters = int(filters + depth_divisor / 2) // depth_divisor * depth_divisor
    new_filters = max(depth_divisor, new_filters)
    # Make sure that round down does not go down by more than 10%.
    if new_filters < 0.9 * filters:
        new_filters += depth_divisor
    return int(new_filters)


def round_repeats(repeats, depth_coefficient):
    """Round number of repeats based on depth multiplier."""

    return int(math.ceil(depth_coefficient * repeats))


def mb_conv_block(inputs, block_args, activation, drop_rate=None, prefix='', ):
    """Mobile Inverted Residual Bottleneck.
    If drop_rate is specified, dropout is added to the end
    """

    has_se = (block_args.se_ratio is not None) and (0 < block_args.se_ratio <= 1)
    bn_axis = 3  # channels last

    # workaround over non working dropout with None in noise_shape in tf.keras
    Dropout = get_dropout()

    # Expansion phase
    filters = block_args.input_filters * block_args.expand_ratio
    if block_args.expand_ratio != 1:
        x = layers.Conv2D(filters, 1,
                          padding='same',
                          use_bias=False,
                          kernel_initializer=CONV_KERNEL_INITIALIZER,
                          name=prefix + 'expand_conv')(inputs)
        x = layers.BatchNormalization(axis=bn_axis, name=prefix + 'expand_bn')(x)
        x = layers.Activation(activation, name=prefix + 'expand_activation')(x)
    else:
        x = inputs

    # Depthwise Convolution
    x = layers.DepthwiseConv2D(block_args.kernel_size,
                               strides=block_args.strides,
                               padding='same',
                               use_bias=False,
                               depthwise_initializer=CONV_KERNEL_INITIALIZER,
                               name=prefix + 'dwconv')(x)
    x = layers.BatchNormalization(axis=bn_axis, name=prefix + 'bn')(x)
    x = layers.Activation(activation, name=prefix + 'activation')(x)

    # Squeeze and Excitation phase
    if has_se:
        num_reduced_filters = max(1, int(
            block_args.input_filters * block_args.se_ratio
        ))
        se_tensor = layers.GlobalAveragePooling2D(name=prefix + 'se_squeeze')(x)

        target_shape = (1, 1, filters) if backend.image_data_format() == 'channels_last' else (filters, 1, 1)
        se_tensor = layers.Reshape(target_shape, name=prefix + 'se_reshape')(se_tensor)
        se_tensor = layers.Conv2D(num_reduced_filters, 1,
                                  activation=activation,
                                  padding='same',
                                  use_bias=True,
                                  kernel_initializer=CONV_KERNEL_INITIALIZER,
                                  name=prefix + 'se_reduce')(se_tensor)
        se_tensor = layers.Conv2D(filters, 1,
                                  activation='sigmoid',
                                  padding='same',
                                  use_bias=True,
                                  kernel_initializer=CONV_KERNEL_INITIALIZER,
                                  name=prefix + 'se_expand')(se_tensor)
        if backend.backend() == 'theano':
            # For the Theano backend, we have to explicitly make
            # the excitation weights broadcastable.
            pattern = ([True, True, True, False] if backend.image_data_format() == 'channels_last'
                       else [True, False, True, True])
            se_tensor = layers.Lambda(
                # pylint: disable=no-member
                lambda x: backend.pattern_broadcast(x, pattern),
                # pylint: enable=no-member
                name=prefix + 'se_broadcast')(se_tensor)  
        x = layers.multiply([x, se_tensor], name=prefix + 'se_excite')

    # Output phase
    x = layers.Conv2D(block_args.output_filters, 1,
                      padding='same',
                      use_bias=False,
                      kernel_initializer=CONV_KERNEL_INITIALIZER,
                      name=prefix + 'project_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, name=prefix + 'project_bn')(x)
    if block_args.id_skip and all(
            s == 1 for s in block_args.strides
    ) and block_args.input_filters == block_args.output_filters:
        if drop_rate and (drop_rate > 0):
            x = Dropout(drop_rate,
                        noise_shape=(None, 1, 1, 1),
                        name=prefix + 'drop')(x)
        x = layers.add([x, inputs], name=prefix + 'add')

    return x

def EfficientNet(width_coefficient,
                 depth_coefficient,
                 default_resolution,
                 dropout_rate=0.2,
                 drop_connect_rate=0.2,
                 depth_divisor=8,
                 blocks_args=DEFAULT_BLOCKS_ARGS,
                 model_name='efficientnet',
                 include_top=True,
                 weights='imagenet',
                 input_tensor=None,
                 input_shape=None,
                 pooling=None,
                 classes=1000,
                 **kwargs):
    """Instantiates the EfficientNet architecture using given scaling coefficients.
    Optionally loads weights pre-trained on ImageNet.
    Note that the data format convention used by the model is
    the one specified in your Keras config at `~/.keras/keras.json`.
    # Arguments
        width_coefficient: float, scaling coefficient for network width.
        depth_coefficient: float, scaling coefficient for network depth.
        default_resolution: int, default input image size.
        dropout_rate: float, dropout rate before final classifier layer.
        drop_connect_rate: float, dropout rate at skip connections.
        depth_divisor: int.
        blocks_args: A list of BlockArgs to construct block modules.
        model_name: string, model name.
        include_top: whether to include the fully-connected
            layer at the top of the network.
        weights: one of `None` (random initialization),
              'imagenet' (pre-training on ImageNet),
              or the path to the weights file to be loaded.
        input_tensor: optional Keras tensor
            (i.e. output of `layers.Input()`)
            to use as image input for the model.
        input_shape: optional shape tuple, only to be specified
            if `include_top` is False.
            It should have exactly 3 inputs channels.
        pooling: optional pooling mode for feature extraction
            when `include_top` is `False`.
            - `None` means that the output of the model will be
                the 4D tensor output of the
                last convolutional layer.
            - `avg` means that global average pooling
                will be applied to the output of the
                last convolutional layer, and thus
                the output of the model will be a 2D tensor.
            - `max` means that global max pooling will
                be applied.
        classes: optional number of classes to classify images
            into, only to be specified if `include_top` is True, and
            if no `weights` argument is specified.
    # Returns
        A Keras model instance.
    # Raises
        ValueError: in case of invalid argument for `weights`,
            or invalid input shape.
    """
    global backend, layers, models, keras_utils

    # Determine proper input shape
    # default 224x224x3 if input_shape=none
    # https://github.com/keras-team/keras-applications/blob/master/keras_applications/imagenet_utils.py
    input_shape = _obtain_input_shape(input_shape,
                                      default_size=default_resolution,
                                      min_size=32,
                                      data_format=backend.image_data_format(),
                                      require_flatten=False,  # changed from include_top in original code
                                      weights=weights)
    input_name = 'input_img'

    if input_tensor is None:
        img_input = layers.Input(shape=input_shape, name=input_name)
    else:
        if backend.backend() == 'tensorflow':
            from tensorflow.python.keras.backend import is_keras_tensor
        else:
            is_keras_tensor = backend.is_keras_tensor
        if not is_keras_tensor(input_tensor):
            img_input = layers.Input(tensor=input_tensor, shape=input_shape, name=input_name)
        else:
            img_input = input_tensor

    bn_axis = 3 if backend.image_data_format() == 'channels_last' else 1
    activation = get_swish(**kwargs)

    stem, stem_output = get_stem(img_input, width_coefficient, depth_divisor, bn_axis, activation)

    blocks, blocks_output = get_blocks(stem_output.shape[1:], blocks_args, width_coefficient, depth_coefficient, depth_divisor, drop_connect_rate, activation)

    # Build top
    # TODO is this in PyTorch version?
    final_part, _ = get_final_part_of_encoder(blocks_output.shape[1:], width_coefficient, depth_divisor, bn_axis, activation)

    # if include_top:  
    assert not include_top

    #     x = final_part_output
    #     # this is NEVER true with Zoobot. 
    #     # `define_model.get_model(include_top=True)` will build my own top, not this. 
    #     # Left for comparison only!

    #     logging.warning('Building EffNet with default top - almost certainly a bad idea for classifying GZ galaxies!')
        
    #     if dropout_rate and dropout_rate > 0:
    #         x = layers.Dropout(dropout_rate, name='top_dropout')(x)
    #         # I use constantly-on dropout instead
    #         # top layer dropout needs to be high to do anything much
    #         # x = custom_layers.PermaDropout(dropout_rate, name='top_dropout')(x)  
    #     x = layers.Dense(classes,
    #                      activation='softmax',
    #                      kernel_initializer=DENSE_KERNEL_INITIALIZER,
    #                      name='probs')(x)


    # Ensure that the model takes into account
    # any potential predecessors of `input_tensor`.
    if input_tensor is not None:
        inputs = keras_utils.get_source_inputs(input_tensor)
    else:
        inputs = img_input

    # Create model
    x = stem(img_input)
    x = blocks(x)
    x = final_part(x)
    model = models.Model(inputs, x, name=model_name)



    # Load weights.
    # if weights == 'imagenet':
    #     logging.warning('Loading pretrained imagenet weights')
    #     if include_top:
    #         file_name = model_name + '_weights_tf_dim_ordering_tf_kernels_autoaugment.h5'
    #         file_hash = IMAGENET_WEIGHTS_HASHES[model_name][0]
    #     else:
    #         file_name = model_name + '_weights_tf_dim_ordering_tf_kernels_autoaugment_notop.h5'
    #         file_hash = IMAGENET_WEIGHTS_HASHES[model_name][1]
    #     weights_path = tf.keras.utils.get_file(
    #         file_name,
    #         IMAGENET_WEIGHTS_PATH + file_name,
    #         cache_subdir='imagenet',
    #         file_hash=file_hash,
    #     )
    #     model.load_weights(weights_path)

    return model


def get_stem(img_input, width_coefficient, depth_divisor, bn_axis, activation):
    # Build stem
    x = img_input
    x = layers.Conv2D(round_filters(32, width_coefficient, depth_divisor), 3,
                      strides=(2, 2),
                      padding='same',
                      use_bias=False,
                      kernel_initializer=CONV_KERNEL_INITIALIZER,
                      name='stem_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, name='stem_bn')(x)
    x = layers.Activation(activation, name='stem_activation')(x)
    stem = tf.keras.Model(inputs=img_input, outputs=x, name='stem')
    return stem, x

def get_blocks(input_shape, blocks_args, width_coefficient, depth_coefficient, depth_divisor, drop_connect_rate, activation):

    block_input = tf.keras.Input(input_shape)
    x = block_input

    # Build blocks
    num_blocks_total = sum(round_repeats(block_args.num_repeat,
                                         depth_coefficient) for block_args in blocks_args)
    block_num = 0
    for idx, block_args in enumerate(blocks_args):
        assert block_args.num_repeat > 0
        # Update block input and output filters based on depth multiplier.
        block_args = block_args._replace(
            input_filters=round_filters(block_args.input_filters,
                                        width_coefficient, depth_divisor),
            output_filters=round_filters(block_args.output_filters,
                                         width_coefficient, depth_divisor),
            num_repeat=round_repeats(block_args.num_repeat, depth_coefficient))

        # The first block needs to take care of stride and filter size increase.
        drop_rate = drop_connect_rate * float(block_num) / num_blocks_total
        x = mb_conv_block(x, block_args,
                          activation=activation,
                          drop_rate=drop_rate,  # actually related to drop_connect_rate
                          prefix='block{}a_'.format(idx + 1))
        block_num += 1
        if block_args.num_repeat > 1:
            # pylint: disable=protected-access
            block_args = block_args._replace(
                input_filters=block_args.output_filters, strides=[1, 1])
            # pylint: enable=protected-access
            for bidx in range(block_args.num_repeat - 1):  # no need for six library, assume user is on python 3
                drop_rate = drop_connect_rate * float(block_num) / num_blocks_total
                block_prefix = 'block{}{}_'.format(
                    idx + 1,
                    string.ascii_lowercase[bidx + 1]
                )
                x = mb_conv_block(x, block_args,
                                  activation=activation,
                                  drop_rate=drop_rate,
                                  prefix=block_prefix)
                block_num += 1

    blocks = tf.keras.Model(inputs=block_input, outputs=x, name='blocks')
    return blocks, x

def get_final_part_of_encoder(input_shape, width_coefficient, depth_divisor, bn_axis, activation):

    final_input = tf.keras.Input(input_shape)
    x = final_input

    x = layers.Conv2D(round_filters(1280, width_coefficient, depth_divisor), 1,
                      padding='same',
                      use_bias=False,
                      kernel_initializer=CONV_KERNEL_INITIALIZER,
                      name='top_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, name='top_bn')(x)
    x = layers.Activation(activation, name='top_activation')(x)

    # moved outside of top, like in pytorch version
    # I feel this is more intuitive - base model is now an encoder with no extra dims
    # global average pooling = adaptive average pooling over all dimensions (except batch)
    x = layers.GlobalAveragePooling2D(name='avg_pool')(x)

    return tf.keras.Model(inputs=final_input, outputs=x, name='final_encoder_part'), x


# layer to replace batchnorm which does nothing and has no trainable parameters
# class BatchNormalizationDummy(tf.keras.layers.Layer):

#     def __init__(self, axis=None, trainable=True, name=None, dtype=None, dynamic=False, **kwargs):
#         super().__init__(trainable, name, dtype, dynamic, **kwargs)

#     def forward(self, x):
#         return x

def EfficientNetB0(
        include_top=True,
        weights=None,
        input_tensor=None,
        input_shape=None,
        pooling=None,
        classes=1000,
        **kwargs
):
    return EfficientNet(
        1.0, 1.0, 224, 0.2,
        model_name='efficientnet-b0',  # do not change model name in order to read correct imagenet weights file
        include_top=include_top, weights=weights,
        input_tensor=input_tensor, input_shape=input_shape,
        pooling=pooling, classes=classes,
        **kwargs
    )


def EfficientNetB3(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   dropout_rate=0.3,  # added here  as it has an explicit default
                   input_size=300,  # 300 by paper design, convenient/hacky to change
                   **kwargs):
    return EfficientNet(
        1.2, 1.4, input_size, dropout_rate,
        model_name='efficientnet-b3',
        include_top=include_top, weights=weights,
        input_tensor=input_tensor, input_shape=input_shape,
        pooling=pooling, classes=classes,
        **kwargs
    )


def EfficientNetB7(
        include_top=True,
        weights='imagenet',
        input_tensor=None,
        input_shape=None,
        pooling=None,
        classes=1000,
        **kwargs
):
    return EfficientNet(
        2.0, 3.1, 600, 0.5,
        model_name='efficientnet-b7',
        include_top=include_top, weights=weights,
        input_tensor=input_tensor, input_shape=input_shape,
        pooling=pooling, classes=classes,
        **kwargs
    )


if __name__ == '__main__':

    model = EfficientNetB0(include_top=False, input_shape=(224, 224, 1))
    print(model.summary())

    """
    Layer (type)                Output Shape              Param #   
    =================================================================
    input_img (InputLayer)      [(None, 224, 224, 1)]     0         
                                                                    
    stem (Functional)           (None, 112, 112, 32)      416       
                                                                    
    blocks (Functional)         (None, 7, 7, 320)         3633852   
                                                                    
    final_encoder_part (Functio  (None, 1280)             414720    
    nal)                                                            
                                                                    
    =================================================================
    Total params: 4,048,988
    Trainable params: 4,006,972  # trainable params match, but pytorch has 0 nontrainable params
    Non-trainable params: 42,016
    _________________________________________________________________

    Without batchnorm:
    Total params: 3,964,956
    Trainable params: 3,964,956
    Non-trainable params: 0
    """

    # print(model.get_layer(name='blocks').summary())
    """
    Model: "blocks"
__________________________________________________________________________________________________
 Layer (type)                   Output Shape         Param #     Connected to                     
==================================================================================================
 input_1 (InputLayer)           [(None, 112, 112, 3  0           []                               
                                2)]                                                               
                                                                                                  
 block1a_dwconv (DepthwiseConv2  (None, 112, 112, 32  288        ['input_1[0][0]']                
 D)                             )                                                                 
                                                                                                  
 block1a_bn (BatchNormalization  (None, 112, 112, 32  128        ['block1a_dwconv[0][0]']         
 )                              )                                                                 
                                                                                                  
 block1a_activation (Activation  (None, 112, 112, 32  0          ['block1a_bn[0][0]']             
 )                              )                                                                 
                                                                                                  
 block1a_se_squeeze (GlobalAver  (None, 32)          0           ['block1a_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block1a_se_reshape (Reshape)   (None, 1, 1, 32)     0           ['block1a_se_squeeze[0][0]']     
                                                                                                  
 block1a_se_reduce (Conv2D)     (None, 1, 1, 8)      264         ['block1a_se_reshape[0][0]']     
                                                                                                  
 block1a_se_expand (Conv2D)     (None, 1, 1, 32)     288         ['block1a_se_reduce[0][0]']      
                                                                                                  
 block1a_se_excite (Multiply)   (None, 112, 112, 32  0           ['block1a_activation[0][0]',     
                                )                                 'block1a_se_expand[0][0]']      
                                                                                                  
 block1a_project_conv (Conv2D)  (None, 112, 112, 16  512         ['block1a_se_excite[0][0]']      
                                )                                                                 
                                                                                                  
 block1a_project_bn (BatchNorma  (None, 112, 112, 16  64         ['block1a_project_conv[0][0]']   
 lization)                      )                                                                 
                                                                                                  
 block2a_expand_conv (Conv2D)   (None, 112, 112, 96  1536        ['block1a_project_bn[0][0]']     
                                )                                                                 
                                                                                                  
 block2a_expand_bn (BatchNormal  (None, 112, 112, 96  384        ['block2a_expand_conv[0][0]']    
 ization)                       )                                                                 
                                                                                                  
 block2a_expand_activation (Act  (None, 112, 112, 96  0          ['block2a_expand_bn[0][0]']      
 ivation)                       )                                                                 
                                                                                                  
 block2a_dwconv (DepthwiseConv2  (None, 56, 56, 96)  864         ['block2a_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block2a_bn (BatchNormalization  (None, 56, 56, 96)  384         ['block2a_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block2a_activation (Activation  (None, 56, 56, 96)  0           ['block2a_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block2a_se_squeeze (GlobalAver  (None, 96)          0           ['block2a_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block2a_se_reshape (Reshape)   (None, 1, 1, 96)     0           ['block2a_se_squeeze[0][0]']     
                                                                                                  
 block2a_se_reduce (Conv2D)     (None, 1, 1, 4)      388         ['block2a_se_reshape[0][0]']     
                                                                                                  
 block2a_se_expand (Conv2D)     (None, 1, 1, 96)     480         ['block2a_se_reduce[0][0]']      
                                                                                                  
 block2a_se_excite (Multiply)   (None, 56, 56, 96)   0           ['block2a_activation[0][0]',     
                                                                  'block2a_se_expand[0][0]']      
                                                                                                  
 block2a_project_conv (Conv2D)  (None, 56, 56, 24)   2304        ['block2a_se_excite[0][0]']      
                                                                                                  
 block2a_project_bn (BatchNorma  (None, 56, 56, 24)  96          ['block2a_project_conv[0][0]']   
 lization)                                                                                        
                                                                                                  
 block2b_expand_conv (Conv2D)   (None, 56, 56, 144)  3456        ['block2a_project_bn[0][0]']     
                                                                                                  
 block2b_expand_bn (BatchNormal  (None, 56, 56, 144)  576        ['block2b_expand_conv[0][0]']    
 ization)                                                                                         
                                                                                                  
 block2b_expand_activation (Act  (None, 56, 56, 144)  0          ['block2b_expand_bn[0][0]']      
 ivation)                                                                                         
                                                                                                  
 block2b_dwconv (DepthwiseConv2  (None, 56, 56, 144)  1296       ['block2b_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block2b_bn (BatchNormalization  (None, 56, 56, 144)  576        ['block2b_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block2b_activation (Activation  (None, 56, 56, 144)  0          ['block2b_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block2b_se_squeeze (GlobalAver  (None, 144)         0           ['block2b_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block2b_se_reshape (Reshape)   (None, 1, 1, 144)    0           ['block2b_se_squeeze[0][0]']     
                                                                                                  
 block2b_se_reduce (Conv2D)     (None, 1, 1, 6)      870         ['block2b_se_reshape[0][0]']     
                                                                                                  
 block2b_se_expand (Conv2D)     (None, 1, 1, 144)    1008        ['block2b_se_reduce[0][0]']      
                                                                                                  
 block2b_se_excite (Multiply)   (None, 56, 56, 144)  0           ['block2b_activation[0][0]',     
                                                                  'block2b_se_expand[0][0]']      
                                                                                                  
 block2b_project_conv (Conv2D)  (None, 56, 56, 24)   3456        ['block2b_se_excite[0][0]']      
                                                                                                  
 block2b_project_bn (BatchNorma  (None, 56, 56, 24)  96          ['block2b_project_conv[0][0]']   
 lization)                                                                                        
                                                                                                  
 block2b_drop (FixedDropout)    (None, 56, 56, 24)   0           ['block2b_project_bn[0][0]']     
                                                                                                  
 block2b_add (Add)              (None, 56, 56, 24)   0           ['block2b_drop[0][0]',           
                                                                  'block2a_project_bn[0][0]']     
                                                                                                  
 block3a_expand_conv (Conv2D)   (None, 56, 56, 144)  3456        ['block2b_add[0][0]']            
                                                                                                  
 block3a_expand_bn (BatchNormal  (None, 56, 56, 144)  576        ['block3a_expand_conv[0][0]']    
 ization)                                                                                         
                                                                                                  
 block3a_expand_activation (Act  (None, 56, 56, 144)  0          ['block3a_expand_bn[0][0]']      
 ivation)                                                                                         
                                                                                                  
 block3a_dwconv (DepthwiseConv2  (None, 28, 28, 144)  3600       ['block3a_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block3a_bn (BatchNormalization  (None, 28, 28, 144)  576        ['block3a_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block3a_activation (Activation  (None, 28, 28, 144)  0          ['block3a_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block3a_se_squeeze (GlobalAver  (None, 144)         0           ['block3a_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block3a_se_reshape (Reshape)   (None, 1, 1, 144)    0           ['block3a_se_squeeze[0][0]']     
                                                                                                  
 block3a_se_reduce (Conv2D)     (None, 1, 1, 6)      870         ['block3a_se_reshape[0][0]']     
                                                                                                  
 block3a_se_expand (Conv2D)     (None, 1, 1, 144)    1008        ['block3a_se_reduce[0][0]']      
                                                                                                  
 block3a_se_excite (Multiply)   (None, 28, 28, 144)  0           ['block3a_activation[0][0]',     
                                                                  'block3a_se_expand[0][0]']      
                                                                                                  
 block3a_project_conv (Conv2D)  (None, 28, 28, 40)   5760        ['block3a_se_excite[0][0]']      
                                                                                                  
 block3a_project_bn (BatchNorma  (None, 28, 28, 40)  160         ['block3a_project_conv[0][0]']   
 lization)                                                                                        
                                                                                                  
 block3b_expand_conv (Conv2D)   (None, 28, 28, 240)  9600        ['block3a_project_bn[0][0]']     
                                                                                                  
 block3b_expand_bn (BatchNormal  (None, 28, 28, 240)  960        ['block3b_expand_conv[0][0]']    
 ization)                                                                                         
                                                                                                  
 block3b_expand_activation (Act  (None, 28, 28, 240)  0          ['block3b_expand_bn[0][0]']      
 ivation)                                                                                         
                                                                                                  
 block3b_dwconv (DepthwiseConv2  (None, 28, 28, 240)  6000       ['block3b_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block3b_bn (BatchNormalization  (None, 28, 28, 240)  960        ['block3b_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block3b_activation (Activation  (None, 28, 28, 240)  0          ['block3b_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block3b_se_squeeze (GlobalAver  (None, 240)         0           ['block3b_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block3b_se_reshape (Reshape)   (None, 1, 1, 240)    0           ['block3b_se_squeeze[0][0]']     
                                                                                                  
 block3b_se_reduce (Conv2D)     (None, 1, 1, 10)     2410        ['block3b_se_reshape[0][0]']     
                                                                                                  
 block3b_se_expand (Conv2D)     (None, 1, 1, 240)    2640        ['block3b_se_reduce[0][0]']      
                                                                                                  
 block3b_se_excite (Multiply)   (None, 28, 28, 240)  0           ['block3b_activation[0][0]',     
                                                                  'block3b_se_expand[0][0]']      
                                                                                                  
 block3b_project_conv (Conv2D)  (None, 28, 28, 40)   9600        ['block3b_se_excite[0][0]']      
                                                                                                  
 block3b_project_bn (BatchNorma  (None, 28, 28, 40)  160         ['block3b_project_conv[0][0]']   
 lization)                                                                                        
                                                                                                  
 block3b_drop (FixedDropout)    (None, 28, 28, 40)   0           ['block3b_project_bn[0][0]']     
                                                                                                  
 block3b_add (Add)              (None, 28, 28, 40)   0           ['block3b_drop[0][0]',           
                                                                  'block3a_project_bn[0][0]']     
                                                                                                  
 block4a_expand_conv (Conv2D)   (None, 28, 28, 240)  9600        ['block3b_add[0][0]']            
                                                                                                  
 block4a_expand_bn (BatchNormal  (None, 28, 28, 240)  960        ['block4a_expand_conv[0][0]']    
 ization)                                                                                         
                                                                                                  
 block4a_expand_activation (Act  (None, 28, 28, 240)  0          ['block4a_expand_bn[0][0]']      
 ivation)                                                                                         
                                                                                                  
 block4a_dwconv (DepthwiseConv2  (None, 14, 14, 240)  2160       ['block4a_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block4a_bn (BatchNormalization  (None, 14, 14, 240)  960        ['block4a_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block4a_activation (Activation  (None, 14, 14, 240)  0          ['block4a_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block4a_se_squeeze (GlobalAver  (None, 240)         0           ['block4a_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block4a_se_reshape (Reshape)   (None, 1, 1, 240)    0           ['block4a_se_squeeze[0][0]']     
                                                                                                  
 block4a_se_reduce (Conv2D)     (None, 1, 1, 10)     2410        ['block4a_se_reshape[0][0]']     
                                                                                                  
 block4a_se_expand (Conv2D)     (None, 1, 1, 240)    2640        ['block4a_se_reduce[0][0]']      
                                                                                                  
 block4a_se_excite (Multiply)   (None, 14, 14, 240)  0           ['block4a_activation[0][0]',     
                                                                  'block4a_se_expand[0][0]']      
                                                                                                  
 block4a_project_conv (Conv2D)  (None, 14, 14, 80)   19200       ['block4a_se_excite[0][0]']      
                                                                                                  
 block4a_project_bn (BatchNorma  (None, 14, 14, 80)  320         ['block4a_project_conv[0][0]']   
 lization)                                                                                        
                                                                                                  
 block4b_expand_conv (Conv2D)   (None, 14, 14, 480)  38400       ['block4a_project_bn[0][0]']     
                                                                                                  
 block4b_expand_bn (BatchNormal  (None, 14, 14, 480)  1920       ['block4b_expand_conv[0][0]']    
 ization)                                                                                         
                                                                                                  
 block4b_expand_activation (Act  (None, 14, 14, 480)  0          ['block4b_expand_bn[0][0]']      
 ivation)                                                                                         
                                                                                                  
 block4b_dwconv (DepthwiseConv2  (None, 14, 14, 480)  4320       ['block4b_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block4b_bn (BatchNormalization  (None, 14, 14, 480)  1920       ['block4b_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block4b_activation (Activation  (None, 14, 14, 480)  0          ['block4b_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block4b_se_squeeze (GlobalAver  (None, 480)         0           ['block4b_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block4b_se_reshape (Reshape)   (None, 1, 1, 480)    0           ['block4b_se_squeeze[0][0]']     
                                                                                                  
 block4b_se_reduce (Conv2D)     (None, 1, 1, 20)     9620        ['block4b_se_reshape[0][0]']     
                                                                                                  
 block4b_se_expand (Conv2D)     (None, 1, 1, 480)    10080       ['block4b_se_reduce[0][0]']      
                                                                                                  
 block4b_se_excite (Multiply)   (None, 14, 14, 480)  0           ['block4b_activation[0][0]',     
                                                                  'block4b_se_expand[0][0]']      
                                                                                                  
 block4b_project_conv (Conv2D)  (None, 14, 14, 80)   38400       ['block4b_se_excite[0][0]']      
                                                                                                  
 block4b_project_bn (BatchNorma  (None, 14, 14, 80)  320         ['block4b_project_conv[0][0]']   
 lization)                                                                                        
                                                                                                  
 block4b_drop (FixedDropout)    (None, 14, 14, 80)   0           ['block4b_project_bn[0][0]']     
                                                                                                  
 block4b_add (Add)              (None, 14, 14, 80)   0           ['block4b_drop[0][0]',           
                                                                  'block4a_project_bn[0][0]']     
                                                                                                  
 block4c_expand_conv (Conv2D)   (None, 14, 14, 480)  38400       ['block4b_add[0][0]']            
                                                                                                  
 block4c_expand_bn (BatchNormal  (None, 14, 14, 480)  1920       ['block4c_expand_conv[0][0]']    
 ization)                                                                                         
                                                                                                  
 block4c_expand_activation (Act  (None, 14, 14, 480)  0          ['block4c_expand_bn[0][0]']      
 ivation)                                                                                         
                                                                                                  
 block4c_dwconv (DepthwiseConv2  (None, 14, 14, 480)  4320       ['block4c_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block4c_bn (BatchNormalization  (None, 14, 14, 480)  1920       ['block4c_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block4c_activation (Activation  (None, 14, 14, 480)  0          ['block4c_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block4c_se_squeeze (GlobalAver  (None, 480)         0           ['block4c_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block4c_se_reshape (Reshape)   (None, 1, 1, 480)    0           ['block4c_se_squeeze[0][0]']     
                                                                                                  
 block4c_se_reduce (Conv2D)     (None, 1, 1, 20)     9620        ['block4c_se_reshape[0][0]']     
                                                                                                  
 block4c_se_expand (Conv2D)     (None, 1, 1, 480)    10080       ['block4c_se_reduce[0][0]']      
                                                                                                  
 block4c_se_excite (Multiply)   (None, 14, 14, 480)  0           ['block4c_activation[0][0]',     
                                                                  'block4c_se_expand[0][0]']      
                                                                                                  
 block4c_project_conv (Conv2D)  (None, 14, 14, 80)   38400       ['block4c_se_excite[0][0]']      
                                                                                                  
 block4c_project_bn (BatchNorma  (None, 14, 14, 80)  320         ['block4c_project_conv[0][0]']   
 lization)                                                                                        
                                                                                                  
 block4c_drop (FixedDropout)    (None, 14, 14, 80)   0           ['block4c_project_bn[0][0]']     
                                                                                                  
 block4c_add (Add)              (None, 14, 14, 80)   0           ['block4c_drop[0][0]',           
                                                                  'block4b_add[0][0]']            
                                                                                                  
 block5a_expand_conv (Conv2D)   (None, 14, 14, 480)  38400       ['block4c_add[0][0]']            
                                                                                                  
 block5a_expand_bn (BatchNormal  (None, 14, 14, 480)  1920       ['block5a_expand_conv[0][0]']    
 ization)                                                                                         
                                                                                                  
 block5a_expand_activation (Act  (None, 14, 14, 480)  0          ['block5a_expand_bn[0][0]']      
 ivation)                                                                                         
                                                                                                  
 block5a_dwconv (DepthwiseConv2  (None, 14, 14, 480)  12000      ['block5a_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block5a_bn (BatchNormalization  (None, 14, 14, 480)  1920       ['block5a_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block5a_activation (Activation  (None, 14, 14, 480)  0          ['block5a_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block5a_se_squeeze (GlobalAver  (None, 480)         0           ['block5a_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block5a_se_reshape (Reshape)   (None, 1, 1, 480)    0           ['block5a_se_squeeze[0][0]']     
                                                                                                  
 block5a_se_reduce (Conv2D)     (None, 1, 1, 20)     9620        ['block5a_se_reshape[0][0]']     
                                                                                                  
 block5a_se_expand (Conv2D)     (None, 1, 1, 480)    10080       ['block5a_se_reduce[0][0]']      
                                                                                                  
 block5a_se_excite (Multiply)   (None, 14, 14, 480)  0           ['block5a_activation[0][0]',     
                                                                  'block5a_se_expand[0][0]']      
                                                                                                  
 block5a_project_conv (Conv2D)  (None, 14, 14, 112)  53760       ['block5a_se_excite[0][0]']      
                                                                                                  
 block5a_project_bn (BatchNorma  (None, 14, 14, 112)  448        ['block5a_project_conv[0][0]']   
 lization)                                                                                        
                                                                                                  
 block5b_expand_conv (Conv2D)   (None, 14, 14, 672)  75264       ['block5a_project_bn[0][0]']     
                                                                                                  
 block5b_expand_bn (BatchNormal  (None, 14, 14, 672)  2688       ['block5b_expand_conv[0][0]']    
 ization)                                                                                         
                                                                                                  
 block5b_expand_activation (Act  (None, 14, 14, 672)  0          ['block5b_expand_bn[0][0]']      
 ivation)                                                                                         
                                                                                                  
 block5b_dwconv (DepthwiseConv2  (None, 14, 14, 672)  16800      ['block5b_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block5b_bn (BatchNormalization  (None, 14, 14, 672)  2688       ['block5b_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block5b_activation (Activation  (None, 14, 14, 672)  0          ['block5b_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block5b_se_squeeze (GlobalAver  (None, 672)         0           ['block5b_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block5b_se_reshape (Reshape)   (None, 1, 1, 672)    0           ['block5b_se_squeeze[0][0]']     
                                                                                                  
 block5b_se_reduce (Conv2D)     (None, 1, 1, 28)     18844       ['block5b_se_reshape[0][0]']     
                                                                                                  
 block5b_se_expand (Conv2D)     (None, 1, 1, 672)    19488       ['block5b_se_reduce[0][0]']      
                                                                                                  
 block5b_se_excite (Multiply)   (None, 14, 14, 672)  0           ['block5b_activation[0][0]',     
                                                                  'block5b_se_expand[0][0]']      
                                                                                                  
 block5b_project_conv (Conv2D)  (None, 14, 14, 112)  75264       ['block5b_se_excite[0][0]']      
                                                                                                  
 block5b_project_bn (BatchNorma  (None, 14, 14, 112)  448        ['block5b_project_conv[0][0]']   
 lization)                                                                                        
                                                                                                  
 block5b_drop (FixedDropout)    (None, 14, 14, 112)  0           ['block5b_project_bn[0][0]']     
                                                                                                  
 block5b_add (Add)              (None, 14, 14, 112)  0           ['block5b_drop[0][0]',           
                                                                  'block5a_project_bn[0][0]']     
                                                                                                  
 block5c_expand_conv (Conv2D)   (None, 14, 14, 672)  75264       ['block5b_add[0][0]']            
                                                                                                  
 block5c_expand_bn (BatchNormal  (None, 14, 14, 672)  2688       ['block5c_expand_conv[0][0]']    
 ization)                                                                                         
                                                                                                  
 block5c_expand_activation (Act  (None, 14, 14, 672)  0          ['block5c_expand_bn[0][0]']      
 ivation)                                                                                         
                                                                                                  
 block5c_dwconv (DepthwiseConv2  (None, 14, 14, 672)  16800      ['block5c_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block5c_bn (BatchNormalization  (None, 14, 14, 672)  2688       ['block5c_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block5c_activation (Activation  (None, 14, 14, 672)  0          ['block5c_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block5c_se_squeeze (GlobalAver  (None, 672)         0           ['block5c_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block5c_se_reshape (Reshape)   (None, 1, 1, 672)    0           ['block5c_se_squeeze[0][0]']     
                                                                                                  
 block5c_se_reduce (Conv2D)     (None, 1, 1, 28)     18844       ['block5c_se_reshape[0][0]']     
                                                                                                  
 block5c_se_expand (Conv2D)     (None, 1, 1, 672)    19488       ['block5c_se_reduce[0][0]']      
                                                                                                  
 block5c_se_excite (Multiply)   (None, 14, 14, 672)  0           ['block5c_activation[0][0]',     
                                                                  'block5c_se_expand[0][0]']      
                                                                                                  
 block5c_project_conv (Conv2D)  (None, 14, 14, 112)  75264       ['block5c_se_excite[0][0]']      
                                                                                                  
 block5c_project_bn (BatchNorma  (None, 14, 14, 112)  448        ['block5c_project_conv[0][0]']   
 lization)                                                                                        
                                                                                                  
 block5c_drop (FixedDropout)    (None, 14, 14, 112)  0           ['block5c_project_bn[0][0]']     
                                                                                                  
 block5c_add (Add)              (None, 14, 14, 112)  0           ['block5c_drop[0][0]',           
                                                                  'block5b_add[0][0]']            
                                                                                                  
 block6a_expand_conv (Conv2D)   (None, 14, 14, 672)  75264       ['block5c_add[0][0]']            
                                                                                                  
 block6a_expand_bn (BatchNormal  (None, 14, 14, 672)  2688       ['block6a_expand_conv[0][0]']    
 ization)                                                                                         
                                                                                                  
 block6a_expand_activation (Act  (None, 14, 14, 672)  0          ['block6a_expand_bn[0][0]']      
 ivation)                                                                                         
                                                                                                  
 block6a_dwconv (DepthwiseConv2  (None, 7, 7, 672)   16800       ['block6a_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block6a_bn (BatchNormalization  (None, 7, 7, 672)   2688        ['block6a_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block6a_activation (Activation  (None, 7, 7, 672)   0           ['block6a_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block6a_se_squeeze (GlobalAver  (None, 672)         0           ['block6a_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block6a_se_reshape (Reshape)   (None, 1, 1, 672)    0           ['block6a_se_squeeze[0][0]']     
                                                                                                  
 block6a_se_reduce (Conv2D)     (None, 1, 1, 28)     18844       ['block6a_se_reshape[0][0]']     
                                                                                                  
 block6a_se_expand (Conv2D)     (None, 1, 1, 672)    19488       ['block6a_se_reduce[0][0]']      
                                                                                                  
 block6a_se_excite (Multiply)   (None, 7, 7, 672)    0           ['block6a_activation[0][0]',     
                                                                  'block6a_se_expand[0][0]']      
                                                                                                  
 block6a_project_conv (Conv2D)  (None, 7, 7, 192)    129024      ['block6a_se_excite[0][0]']      
                                                                                                  
 block6a_project_bn (BatchNorma  (None, 7, 7, 192)   768         ['block6a_project_conv[0][0]']   
 lization)                                                                                        
                                                                                                  
 block6b_expand_conv (Conv2D)   (None, 7, 7, 1152)   221184      ['block6a_project_bn[0][0]']     
                                                                                                  
 block6b_expand_bn (BatchNormal  (None, 7, 7, 1152)  4608        ['block6b_expand_conv[0][0]']    
 ization)                                                                                         
                                                                                                  
 block6b_expand_activation (Act  (None, 7, 7, 1152)  0           ['block6b_expand_bn[0][0]']      
 ivation)                                                                                         
                                                                                                  
 block6b_dwconv (DepthwiseConv2  (None, 7, 7, 1152)  28800       ['block6b_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block6b_bn (BatchNormalization  (None, 7, 7, 1152)  4608        ['block6b_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block6b_activation (Activation  (None, 7, 7, 1152)  0           ['block6b_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block6b_se_squeeze (GlobalAver  (None, 1152)        0           ['block6b_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block6b_se_reshape (Reshape)   (None, 1, 1, 1152)   0           ['block6b_se_squeeze[0][0]']     
                                                                                                  
 block6b_se_reduce (Conv2D)     (None, 1, 1, 48)     55344       ['block6b_se_reshape[0][0]']     
                                                                                                  
 block6b_se_expand (Conv2D)     (None, 1, 1, 1152)   56448       ['block6b_se_reduce[0][0]']      
                                                                                                  
 block6b_se_excite (Multiply)   (None, 7, 7, 1152)   0           ['block6b_activation[0][0]',     
                                                                  'block6b_se_expand[0][0]']      
                                                                                                  
 block6b_project_conv (Conv2D)  (None, 7, 7, 192)    221184      ['block6b_se_excite[0][0]']      
                                                                                                  
 block6b_project_bn (BatchNorma  (None, 7, 7, 192)   768         ['block6b_project_conv[0][0]']   
 lization)                                                                                        
                                                                                                  
 block6b_drop (FixedDropout)    (None, 7, 7, 192)    0           ['block6b_project_bn[0][0]']     
                                                                                                  
 block6b_add (Add)              (None, 7, 7, 192)    0           ['block6b_drop[0][0]',           
                                                                  'block6a_project_bn[0][0]']     
                                                                                                  
 block6c_expand_conv (Conv2D)   (None, 7, 7, 1152)   221184      ['block6b_add[0][0]']            
                                                                                                  
 block6c_expand_bn (BatchNormal  (None, 7, 7, 1152)  4608        ['block6c_expand_conv[0][0]']    
 ization)                                                                                         
                                                                                                  
 block6c_expand_activation (Act  (None, 7, 7, 1152)  0           ['block6c_expand_bn[0][0]']      
 ivation)                                                                                         
                                                                                                  
 block6c_dwconv (DepthwiseConv2  (None, 7, 7, 1152)  28800       ['block6c_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block6c_bn (BatchNormalization  (None, 7, 7, 1152)  4608        ['block6c_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block6c_activation (Activation  (None, 7, 7, 1152)  0           ['block6c_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block6c_se_squeeze (GlobalAver  (None, 1152)        0           ['block6c_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block6c_se_reshape (Reshape)   (None, 1, 1, 1152)   0           ['block6c_se_squeeze[0][0]']     
                                                                                                  
 block6c_se_reduce (Conv2D)     (None, 1, 1, 48)     55344       ['block6c_se_reshape[0][0]']     
                                                                                                  
 block6c_se_expand (Conv2D)     (None, 1, 1, 1152)   56448       ['block6c_se_reduce[0][0]']      
                                                                                                  
 block6c_se_excite (Multiply)   (None, 7, 7, 1152)   0           ['block6c_activation[0][0]',     
                                                                  'block6c_se_expand[0][0]']      
                                                                                                  
 block6c_project_conv (Conv2D)  (None, 7, 7, 192)    221184      ['block6c_se_excite[0][0]']      
                                                                                                  
 block6c_project_bn (BatchNorma  (None, 7, 7, 192)   768         ['block6c_project_conv[0][0]']   
 lization)                                                                                        
                                                                                                  
 block6c_drop (FixedDropout)    (None, 7, 7, 192)    0           ['block6c_project_bn[0][0]']     
                                                                                                  
 block6c_add (Add)              (None, 7, 7, 192)    0           ['block6c_drop[0][0]',           
                                                                  'block6b_add[0][0]']            
                                                                                                  
 block6d_expand_conv (Conv2D)   (None, 7, 7, 1152)   221184      ['block6c_add[0][0]']            
                                                                                                  
 block6d_expand_bn (BatchNormal  (None, 7, 7, 1152)  4608        ['block6d_expand_conv[0][0]']    
 ization)                                                                                         
                                                                                                  
 block6d_expand_activation (Act  (None, 7, 7, 1152)  0           ['block6d_expand_bn[0][0]']      
 ivation)                                                                                         
                                                                                                  
 block6d_dwconv (DepthwiseConv2  (None, 7, 7, 1152)  28800       ['block6d_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block6d_bn (BatchNormalization  (None, 7, 7, 1152)  4608        ['block6d_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block6d_activation (Activation  (None, 7, 7, 1152)  0           ['block6d_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block6d_se_squeeze (GlobalAver  (None, 1152)        0           ['block6d_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block6d_se_reshape (Reshape)   (None, 1, 1, 1152)   0           ['block6d_se_squeeze[0][0]']     
                                                                                                  
 block6d_se_reduce (Conv2D)     (None, 1, 1, 48)     55344       ['block6d_se_reshape[0][0]']     
                                                                                                  
 block6d_se_expand (Conv2D)     (None, 1, 1, 1152)   56448       ['block6d_se_reduce[0][0]']      
                                                                                                  
 block6d_se_excite (Multiply)   (None, 7, 7, 1152)   0           ['block6d_activation[0][0]',     
                                                                  'block6d_se_expand[0][0]']      
                                                                                                  
 block6d_project_conv (Conv2D)  (None, 7, 7, 192)    221184      ['block6d_se_excite[0][0]']      
                                                                                                  
 block6d_project_bn (BatchNorma  (None, 7, 7, 192)   768         ['block6d_project_conv[0][0]']   
 lization)                                                                                        
                                                                                                  
 block6d_drop (FixedDropout)    (None, 7, 7, 192)    0           ['block6d_project_bn[0][0]']     
                                                                                                  
 block6d_add (Add)              (None, 7, 7, 192)    0           ['block6d_drop[0][0]',           
                                                                  'block6c_add[0][0]']            
                                                                                                  
 block7a_expand_conv (Conv2D)   (None, 7, 7, 1152)   221184      ['block6d_add[0][0]']            
                                                                                                  
 block7a_expand_bn (BatchNormal  (None, 7, 7, 1152)  4608        ['block7a_expand_conv[0][0]']    
 ization)                                                                                         
                                                                                                  
 block7a_expand_activation (Act  (None, 7, 7, 1152)  0           ['block7a_expand_bn[0][0]']      
 ivation)                                                                                         
                                                                                                  
 block7a_dwconv (DepthwiseConv2  (None, 7, 7, 1152)  10368       ['block7a_expand_activation[0][0]
 D)                                                              ']                               
                                                                                                  
 block7a_bn (BatchNormalization  (None, 7, 7, 1152)  4608        ['block7a_dwconv[0][0]']         
 )                                                                                                
                                                                                                  
 block7a_activation (Activation  (None, 7, 7, 1152)  0           ['block7a_bn[0][0]']             
 )                                                                                                
                                                                                                  
 block7a_se_squeeze (GlobalAver  (None, 1152)        0           ['block7a_activation[0][0]']     
 agePooling2D)                                                                                    
                                                                                                  
 block7a_se_reshape (Reshape)   (None, 1, 1, 1152)   0           ['block7a_se_squeeze[0][0]']     
                                                                                                  
 block7a_se_reduce (Conv2D)     (None, 1, 1, 48)     55344       ['block7a_se_reshape[0][0]']     
                                                                                                  
 block7a_se_expand (Conv2D)     (None, 1, 1, 1152)   56448       ['block7a_se_reduce[0][0]']      
                                                                                                  
 block7a_se_excite (Multiply)   (None, 7, 7, 1152)   0           ['block7a_activation[0][0]',     
                                                                  'block7a_se_expand[0][0]']      
                                                                                                  
 block7a_project_conv (Conv2D)  (None, 7, 7, 320)    368640      ['block7a_se_excite[0][0]']      
                                                                                                  
 block7a_project_bn (BatchNorma  (None, 7, 7, 320)   1280        ['block7a_project_conv[0][0]']   
 lization)                                                                                        
    """

    print(model.get_layer(name='final_encoder_part').summary())
    """
    Model: "final_encoder_part"
    _________________________________________________________________
    Layer (type)                Output Shape              Param #   
    =================================================================
    input_2 (InputLayer)        [(None, 7, 7, 320)]       0         
                                                                    
    top_conv (Conv2D)           (None, 7, 7, 1280)        409600    
                                                                    
    top_bn (BatchNormalization)  (None, 7, 7, 1280)       5120      
                                                                    
    top_activation (Activation)  (None, 7, 7, 1280)       0         
                                                                    
    avg_pool (GlobalAveragePool  (None, 1280)             0   
    
    shapes match pytorch
    avg_pool rather than *adaptive* avg_pool - possibly just adaptive as in any input shape?
    batchnorm listed as twice as many params in TF (5120 vs 2560)

    There are 4 params per input feature in batch norm: 
        beta and gamma (momentum), trainable
        mean and variance (statistics), which update but are not directly optimised
    I think torchsummary considers statistics non-trainable while keras.Summary considers trainable
    """


import logging

import tensorflow as tf

from zoobot.tensorflow.estimators import efficientnet_standard, efficientnet_custom, custom_layers

# https://www.tensorflow.org/api_docs/python/tf/keras/callbacks/TensorBoard
# for Functional, must be wrapped in Lambda layer
class LogHistogram(tf.keras.layers.Layer):
    def __init__(self, name, description=None):
        super(LogHistogram, self).__init__()
        self.log_name = name  # 'name' is taken
        self.description = description
    def call(self, data, training):
        if training:
            tf.summary.histogram(name=self.log_name, data=data, description=self.description)
        return data
class LogScalar(tf.keras.layers.Layer):
    def __init__(self, name, description=None):
        super(LogScalar, self).__init__()
        self.log_name = name  # 'name' is taken
        self.description = description
    def call(self, data, training):
        if training:
            tf.summary.scalar(name=self.log_name, data=data, description=self.description)
        return data
class LogImage(tf.keras.layers.Layer):
    def __init__(self, name, description=None):
        super(LogImage, self).__init__()
        self.log_name = name  # 'name' is taken
        self.description = description
    def call(self, data, training):
        if training:
            tf.summary.image(name=self.log_name, data=data, description=self.description)
        return data



def get_model(
    output_dim,
    input_size,
    weights_loc=None,
    include_top=True,
    expect_partial=False,
    channels=1,
    use_imagenet_weights=False,
    dropout_rate=0.2,
    test_time_dropout=True,
    get_effnet=efficientnet_standard.EfficientNetB0
    ):
    """
    Create a trainable efficientnet model.
    First layers are galaxy-appropriate augmentation layers - see :meth:`zoobot.estimators.define_model.add_augmentation_layers`.
    Expects single channel image e.g. (300, 300, 1), likely with leading batch dimension.

    Optionally (by default) include the head (output layers) used for GZ DECaLS.
    Specifically, global average pooling followed by a dense layer suitable for predicting dirichlet parameters.
    See ``efficientnet_custom.custom_top_dirichlet``

    Args:
        output_dim (int): Dimension of head dense layer. No effect when include_top=False.
        input_size (int): Length of initial image e.g. 300 (assumed square)
        weights_loc (str, optional): If str, load weights from efficientnet checkpoint at this location. Defaults to None.
        include_top (bool, optional): If True, include head used for GZ DECaLS: global pooling and dense layer. Defaults to True.
        expect_partial (bool, optional): If True, do not raise partial match error when loading weights (likely for optimizer state). Defaults to False.
        channels (int, default 1): Number of channels i.e. C in NHWC-dimension inputs. 

    Returns:
        tf.keras.Model: trainable efficientnet model including augmentations and optional head
    """
    logging.info('Input size {} (should match resize_after_crop)'.format(
        input_size))

    # model = CustomSequential()  # to log the input image for debugging
    # model = tf.keras.Sequential()

    input_shape = (input_size, input_size, channels)

    inputs = tf.keras.Input(shape=input_shape, name='preprocessed_image_batch')

    x = LogImage(
        name='images_as_input',
        description='Images passed to Zoobot'
    )(inputs)

    # now headless
    effnet = efficientnet_custom.define_headless_efficientnet(
        input_shape=input_shape,
        get_effnet=get_effnet,
        # further kwargs will be passed to get_effnet
        use_imagenet_weights=use_imagenet_weights,
    )
    x = effnet(x)
    x = LogHistogram(name='embedding')(x)

    # Functional head
    if include_top:
        assert output_dim is not None
        if test_time_dropout:
            logging.info('Using test-time dropout')
            dropout_layer = custom_layers.PermaDropout
        else:
            logging.info('Not using test-time dropout')
            dropout_layer = tf.keras.layers.Dropout
        x = dropout_layer(dropout_rate, name='top_dropout')(x)
        x = efficientnet_custom.custom_top_dirichlet(output_dim)(x)
        x = LogHistogram(name='dirichlet_outputs')(x)

    model = tf.keras.Model(inputs=inputs, outputs=x, name="zoobot")
 
    if weights_loc:
        load_weights(model, weights_loc, expect_partial=expect_partial)

    return model


# inplace
def load_weights(model, checkpoint_loc, expect_partial=False):
    """
    Load weights checkpoint to ``model``.
    Acts inplace i.e. model is modified by reference.

    Args:
        model (tf.keras.Model): Model into which to load checkpoint
        checkpoint_loc (str): path to checkpoint e.g. /path/checkpoints/checkpoint (where checkpoints includes checkpoint.index etc)
        expect_partial (bool, optional): If True, do not raise partial match error when loading weights (likely for optimizer state). Defaults to False.
    """
    # https://www.tensorflow.org/api_docs/python/tf/train/Checkpoint
    logging.info('Loading weights from {}'.format(checkpoint_loc))
    load_status = model.load_weights(checkpoint_loc)
    load_status.assert_nontrivial_match()
    if expect_partial:  # some checkpointed values not in the current program won't match (the optimiser state during predictions, hopefully)
        load_status.expect_partial()
    # everything in the current program should match
    # do after load_status.expect_partial to silence optimizer warnings
    load_status.assert_existing_objects_matched()


def load_model(checkpoint_loc, include_top, input_size, output_dim=34, expect_partial=False, channels=1, dropout_rate=0.2):
    """    
    Utility wrapper for the common task of defining the GZ DECaLS model and then loading a pretrained checkpoint.
    crop_size must match the pretrained model used.
    output_dim must match if ``include_top=True``
    ``input_size`` and ``crop_size`` can vary as image will be resized anyway, but be careful deviating from training procedure.

    Args:
        checkpoint_loc (str): path to checkpoint e.g. /path/checkpoints/checkpoint (where checkpoints includes checkpoint.index etc)
        include_top (bool, optional): If True, include head used for GZ DECaLS: global pooling and dense layer.
        input_size (int): Length of initial image e.g. 300 (assumed square)
        crop_size (int): Length to randomly crop image. See ``get_augmentation_layers``.
        output_dim (int, optional): Dimension of head dense layer. No effect when include_top=False. Defaults to 34.
        expect_partial (bool, optional): If True, do not raise partial match error when loading weights (likely for optimizer state). Defaults to False.

    Returns:
        tf.keras.Model: GZ DECaLS-like model with weights loaded from ``checkpoint_loc``, optionally including GZ DECaLS-like head.
    """

    model = get_model(
        output_dim=output_dim,
        input_size=input_size,
        include_top=include_top,
        channels=channels,
        dropout_rate=dropout_rate
    )
    load_weights(model, checkpoint_loc, expect_partial=expect_partial)
    return model



import logging
from typing import List

import numpy as np

from galaxy_datasets.shared import label_metadata


class Question():

    def __init__(self, question_text:str, answer_text: List, label_cols:List):
        """
        Class representing decision tree question.
        Requires ``label_cols`` as an input in order to find the index (vs. all questions and answers) of this question and each answer.

        Args:
            question_text (str): e.g. 'smooth-or-featured'
            answer_text (List): e.g. ['smooth', 'featured-or-disk]
            label_cols (List): list of all questions and answers e.g. ['smooth-or-featured_smooth', 'smooth-or-featured_featured-or-disk', ...]
        """
        self.text = question_text
        self.answers = create_answers(self, answer_text, label_cols)  # passing a reference to self, will stay up-to-date

        self.start_index = min(a.index for a in self.answers)
        self.end_index = max(a.index for a in self.answers)
        assert [self.start_index <= a.index <= self.end_index for a in self.answers]

        self._asked_after = None

    @property
    def asked_after(self):
        return self._asked_after

    def __repr__(self):
        return f'{self.text}, indices {self.start_index} to {self.end_index}, asked after {self.asked_after}'


class Answer():

    def __init__(self, text, question, index):
        """
        Class representing decision tree answer.

        Each answer includes the answer text (often used as a column header),
        the corresponding question, and its index in ``label_cols`` (often used for slicing model outputs)

        Args:
            text (str): e.g. 'smooth-or-featured_smooth'
            question (Question): Question class for this answer
            index (int): index of answer in label_cols (0-33 for DECaLS)
        """
        self.text = text
        self.question = question

        self.index = index
        self._next_question = None

    @property
    def next_question(self):
        """
        
        Returns:
            Question: question that follows after this answer. None initially, added by ``set_dependancies``.
        """
        return self._next_question

    def __repr__(self):
        return f'{self.text}, index {self.index}'

    @property
    def pretty_text(self):
        """
        Returns:
            str: Nicely formatted text for plots etc
        """
        return self.text.replace('-',' ').replace('_', ' ').title()


def create_answers(question:Question, answers_texts:List, label_cols:List) -> List[Answer]:
    """
    Instantiate the Answer classes for a given ``Question``.
    Each answer includes the answer text (often used as a column header),
    the corresponding question, and its index in ``label_cols`` (often used for slicing model outputs)

    Args:
        question (Question): question to which to create answers e.g. Question(smooth-or-featured)
        answers_texts (List): answer strings e.g. ['smooth-or-featured_smooth', 'smooth-or-featured_featured-or-disk']
        label_cols (List): list of all questions and answers e.g. ['smooth-or-featured_smooth', 'smooth-or-featured_featured-or-disk', ...]

    Returns:
        List: of Answers to that question e.g. [Answer(smooth-or-featured_smooth), Answer(smooth-or-featured_featured-or-disk)]
    """
    question_text = question.text
    answers = []
    for answer_text in answers_texts:
        answers.append(
            Answer(
                text=question_text + answer_text,  # e.g. smooth-or-featured_smooth
                question=question,
                index=label_cols.index(question_text + answer_text)  # will hopefully throw value error if not found?
                # _next_question not set, set later with dependancies
            )
        )
    return answers
    

def set_dependencies(questions, dependencies):
    """
    Link each answer to question which follows, and vica versa.
    Acts inplace.

    Specifically, for every answer in every question, set answer._next question to refer to the Question which follows that answer.
    Then for that Question, set question._asked_after to that answer.

    Args:
        questions (List): of questions e.g. [Question('smooth-or-featured'), Question('edge-on-disk')]
        dependencies (dict): dict mapping each question (e.g. disk-edge-on) to the answer on which it depends (e.g. smooth-or-featured_featured-or-disk)
    """

    for question in questions:
        prev_answer_text = dependencies[question.text]
        if prev_answer_text is not None:
            try:
                # look through every answer, find those with the same text as "prev answer text" - will be exactly one match
                prev_answer = [a for q in questions for a in q.answers if a.text == prev_answer_text][0] 
            except IndexError:
                raise ValueError(f'{prev_answer_text} not found in dependencies')
            prev_answer._next_question = question
            question._asked_after = prev_answer


class Schema():
    
    def __init__(self, question_answer_pairs:dict, dependencies: dict):
        """
        Relate the df label columns tor question/answer groups and to tfrecod label indices
        Requires that labels be continguous by question - easily satisfied

        Be careful with dependencies:
        - first entry should be the first answer to that question, by df column order
        - second entry should be the last answer to that question, similarly
        - answers in between will be included: these are used to slice
        - df columns must be contigious by question (e.g. not smooth_yes, bar_no, smooth_no) for this to work!

        The following schemas are available via the module (e.g. `from zoobot.shared.schemas import decals_dr5_ortho_schema`):
        - decals_dr5_ortho_schema
        - decals_dr8_ortho_schema
        - decals_all_campaigns_ortho_schema
        - gz2_ortho_schema
        - gz_candels_ortho_schema
        - gz_hubble_ortho_schema
        - cosmic_dawn_ortho_schema
        - cosmic_dawn_schema
        - gz_rings_schema
        - desi_schema
        - gz_evo_v1_schema (this is the schema currently used for pretraining)
        - gz_ukidss_schema
        - gz_jwst_schema

        "ortho" refers to the orthogonal question suffix (-cd, -dr8, etc).

        Args:
            question_answer_pairs (dict): e.g. {'smooth-or-featured: ['_smooth, _featured-or-disk, ...], ...}
            dependencies (dict): dict mapping each question (e.g. disk-edge-on) to the answer on which it depends (e.g. smooth-or-featured_featured-or-disk)
        """
        self.question_answer_pairs = question_answer_pairs
        _, self.label_cols = label_metadata.extract_questions_and_label_cols(question_answer_pairs)
        self.dependencies = dependencies

        self.questions = [Question(question_text, answers_text, self.label_cols) for question_text, answers_text in question_answer_pairs.items()]
        if len(self.questions) > 1:
            set_dependencies(self.questions, self.dependencies)

        assert len(self.question_index_groups) > 0
        assert len(self.questions) == len(self.question_index_groups)


    def get_answer(self, answer_text):
        """

        Args:
            answer_text (str): e.g. 'smooth-or-featured_smooth'

        Raises:
            ValueError: No answer with that answer_text found

        Returns:
            Answer: the answer with matching answer_text e.g. Answer('smooth-or-featured_smooth')
        """
        try:
            return [a for q in self.questions for a in q.answers if a.text == answer_text][0]  # will be exactly one match
        except IndexError:
            raise ValueError('Answer not found: ', answer_text)


    def get_question(self, question_text):
        """

        Args:
            question_text (str): e.g. 'smooth-or-featured'

        Raises:
            ValueError: No question with that question_text found

        Returns:
            Question: the question with matching question_text e.g. Question('smooth-or-featured')
        """
        try:
            return [q for q in self.questions if q.text == question_text][0]  # will be exactly one match
        except  IndexError:
            raise ValueError('Question not found: ', question_text)
    

    @property
    def question_index_groups(self):
        """

        Returns:
            Paired (tuple) integers of (first, last) indices of answers to each question, listed for all questions.
            Useful for slicing model predictions by question.
        """
         # start and end indices of answers to each question in label_cols e.g. [[0, 1]. [1, 3]] 
        return [(q.start_index, q.end_index) for q in self.questions]


    @property
    def named_index_groups(self):
        """
        
        Returns:
            dict: mapping each question to the start and end index of its answers in label_cols, e.g. {Question('smooth-or-featured'): [0, 2], ...}
        """
        return dict(zip(self.questions, self.question_index_groups))


    def joint_p(self, prob_of_answers, answer_text):
        """
        Probability of the answer with ``answer_text`` being asked, given the (predicted) probability of every answer.
        Useful for estimating the relevance of an answer e.g. to ignore predictons for answers less than 50% likely to be asked.

        Broadcasts over batch dimension.

        Args:
            prob_of_answers (np.ndarray): prob. of each answer being asked, of shape (galaxies, answers) where the index of answers matches label_cols
            answer_text (str): which answer to find the prob. of being asked e.g. 'edge-on-disk_yes'

        Returns:
            np.ndarray: prob of that answer being asked e.g. 0.5 for 'edge-on-disk_yes' of prob of 'smooth-or-featured_featured-or-disk' is 0.5. Shape (galaxies).
        """
        assert prob_of_answers.ndim == 2  # batch, p. No 'per model', marginalise first
        # prob(answer) = p(that answer|that q asked) * p(that q_asked) i.e...
        # prob(answer) = p(that answer|that q asked) * p(answer before that q)
        answer = self.get_answer(answer_text)
        p_answer_given_question = prob_of_answers[:, answer.index]
        if all(np.isnan(p_answer_given_question)):
            logging.warning(f'All p_answer_given_question for {answer_text} ({answer.index}) are nan i.e. all fractions are nan - check that labels for this question are appropriate')

        question = answer.question
        prev_answer = question.asked_after
        if prev_answer is None:
            return p_answer_given_question
        else:
            p_prev_answer = self.joint_p(prob_of_answers, prev_answer.text)  # recursive
            return p_answer_given_question * p_prev_answer


    @property
    def answers(self):
        """

        Returns:
            list: all answers
        """
        answers = []
        for q in self.questions:
            for a in q.answers:
                answers.append(a)
        return answers


# and define each schema here, for convenience
decals_dr5_ortho_schema = Schema(label_metadata.decals_dr5_ortho_pairs , label_metadata.decals_ortho_dependencies)
decals_dr8_ortho_schema = Schema(label_metadata.decals_dr8_ortho_pairs , label_metadata.decals_ortho_dependencies)
decals_all_campaigns_ortho_schema = Schema(label_metadata.decals_all_campaigns_ortho_pairs , label_metadata.decals_ortho_dependencies)
gz2_ortho_schema = Schema(label_metadata.gz2_ortho_pairs , label_metadata.gz2_ortho_dependencies)
gz_candels_ortho_schema = Schema(label_metadata.candels_ortho_pairs, label_metadata.candels_ortho_dependencies)
gz_hubble_ortho_schema = Schema(label_metadata.hubble_ortho_pairs, label_metadata.hubble_ortho_dependencies)
cosmic_dawn_ortho_schema = Schema(label_metadata.cosmic_dawn_ortho_pairs , label_metadata.cosmic_dawn_ortho_dependencies)

# schemas without orthogonal question suffix (-cd, -dr8, etc)
cosmic_dawn_schema = Schema(label_metadata.cosmic_dawn_pairs , label_metadata.cosmic_dawn_dependencies)
gz_rings_schema = Schema(label_metadata.rings_pairs, label_metadata.rings_dependencies)
desi_schema = Schema(label_metadata.desi_pairs, label_metadata.desi_dependencies)  # for DESI data release prediction users, not for ML training - no -dr5, -dr8, etc
# note that as this is a call to Schema (and Question and Answer), any logging within those will 
# trigger basicConfig() and prevent user setting their own logging.
# so don't log anything during Schema.__init__!

gz_evo_v1_schema = Schema(label_metadata.gz_evo_v1_pairs, label_metadata.gz_evo_v1_dependencies)

gz_ukidss_schema = Schema(label_metadata.ukidss_ortho_pairs, label_metadata.ukidss_ortho_dependencies)
gz_jwst_schema = Schema(label_metadata.jwst_ortho_pairs, label_metadata.jwst_ortho_dependencies)



import logging
import os


# from lightning_lite.plugins.environments import SLURMEnvironment

# # https://pytorch-lightning.readthedocs.io/en/stable/_modules/lightning_lite/plugins/environments/slurm.html#SLURMEnvironment
# # https://github.com/Lightning-AI/lightning/blob/9c20cad40e4142f8a5e945fe26e919e598f2bd56/src/lightning_lite/plugins/environments/slurm.py
# class ManchesterEnvironment(SLURMEnvironment):

#     def __init__(self, auto_requeue: bool = True, requeue_signal=None) -> None:
#         logging.info('Using Manchester SLURM environment')
#         super().__init__(auto_requeue, requeue_signal)

#     # @staticmethod
#     # def resolve_root_node_address(nodes: str) -> str:
#     #     root_node_address = super().resolve_root_node_address(nodes)
#     #     logging.info(f'root_node_address: {root_node_address}')
#     #     return root_node_address

#     @staticmethod
#     def detect() -> bool:
#         return True
        
#     @property
#     def main_port(self) -> int:
#         main_port = super().main_port
#         logging.info(f'main_port: {main_port}')
#         return main_port
#         # MASTER_PORT will override


from pytorch_lightning.plugins.environments import SLURMEnvironment
class GalahadEnvironment(SLURMEnvironment):
    def __init__(self, **kwargs):
        ntasks_per_node = os.environ["SLURM_TASKS_PER_NODE"].split("(")[0]
        os.environ["SLURM_NTASKS_PER_NODE"] = ntasks_per_node
        logging.warning(f'Within custom slurm environment, --n-tasks-per-node={ntasks_per_node}')
        # os.environ["SLURM_NTASKS"] = str(os.environ["SLURM_NTASKS_PER_NODE"])
        super().__init__(**kwargs)
        self.nnodes = int(os.environ["SLURM_NNODES"])


# if __name__ == '__main__':

#     logging.basicConfig(level=logging.INFO)

#     # slurm_nodelist = "compute-0-[0,9]" # 0,9 works
#     slurm_nodelist = "compute-0-[0,11]"  # 0,11 hangs
#     # 70017 8-9 works

#     env = GalahadEnvironment()
#     root = env.resolve_root_node_address(slurm_nodelist)
#     print(root)

#     print(env.detect())

#     print(env.main_port)


import os
import glob
import logging
import time
import datetime
from typing import List
from pathlib import Path

import numpy as np
import tensorflow as tf

from zoobot.shared import save_predictions


def predict(ds: tf.data.Dataset, model: tf.keras.Model, n_samples: int, label_cols: List, save_loc: str):
    """
    Make and save predictions by model on image dataset.

    Args:
        ds (tf.data.Dataset): dataset yielding batches of (images, id_strs). Preprocessing already applied. id_strs may be the original path to image, or any other galaxy identifier.
        model (tf.keras.Model): trained model with which to make predictions
        n_samples (int): number of repeat predictions. Useful to marginalise over augmentations or MC Dropout.
        label_cols (list): Semantic labels for final model output dimension (e.g. ["smooth", "bar", "merger"]). Only used for output csv/hdf5 notes.
        save_loc (str): path to save csv of predictions.
    """

    # to make sure images and id_str line up, load id_str back out from dataset
    id_str_ds = ds.map(lambda _, id_str: id_str)
    try:
        image_id_strs = [id_str.numpy().decode('utf-8') for id_str_batch in id_str_ds for id_str in id_str_batch]
    except AttributeError:
        raise AttributeError(
            'Dataset does not appear to be yielding batches of (images, id strings). \
            Check with dataset.take(1). Make sure you create the dataset with include_id_str=True and labels=None'
        )

    logging.info('Beginning predictions')
    start = datetime.datetime.fromtimestamp(time.time())
    logging.info('Starting at: {}'.format(start.strftime('%Y-%m-%d %H:%M:%S')))

    predictions = np.stack([model.predict(ds) for n in range(n_samples)], axis=-1)
    logging.info('Predictions complete - {}'.format(predictions.shape))

    if save_loc.endswith('.csv'):      # save as pandas df
        save_predictions.predictions_to_csv(predictions, image_id_strs, label_cols, save_loc)
    elif save_loc.endswith('.hdf5'):
        save_predictions.predictions_to_hdf5(predictions, image_id_strs, label_cols, save_loc)
    else:
        logging.warning('Save format of {} not recognised - assuming csv'.format(save_loc))
        save_predictions.predictions_to_csv(predictions, image_id_strs, label_cols, save_loc)

    logging.info(f'Predictions saved to {save_loc}')

    end = datetime.datetime.fromtimestamp(time.time())
    logging.info('Completed at: {}'.format(end.strftime('%Y-%m-%d %H:%M:%S')))
    logging.info('Time elapsed: {}'.format(end - start))



def paths_in_folder(folder: str, file_format: str, recursive=False):
    """
    Find all files of ``file_format`` in ``folder``, optionally recursively.
    Useful for getting list of image paths.

    Args:
        folder (str): path to folder to search
        file_format (str): include only files ending with this file format
        recursive (bool, optional): If True, search recursively i.e. all subfolders. Defaults to False.

    Returns:
        list: relative paths to all files of ``file_format`` in ``folder``.

    """
    assert os.path.isdir(folder)
    if recursive:  # in all subfolders, recursively
        unordered_paths = [str(path) for path in Path(folder).rglob('*.{}'.format(file_format))]
    else:
        unordered_paths = list(glob.glob('{}/*.{}'.format(folder, file_format)))  # only in that folder
    return unordered_paths



from typing import Tuple
import logging

import torch
import pyro


def calculate_multiquestion_loss(labels: torch.Tensor, predictions: torch.Tensor, question_index_groups: Tuple, careful=True) -> torch.Tensor:
    """
    The full decision tree loss used for training GZ DECaLS models

    Negative log likelihood of observing ``labels`` (volunteer answers to all questions)
    from Dirichlet-Multinomial distributions for each question, using concentrations ``predictions``.
    
    Args:
        labels (torch.Tensor): (galaxy, k successes) where k successes dimension is indexed by question_index_groups.
        predictions (torch.Tensor):  Dirichlet concentrations, matching shape of labels
        question_index_groups (list): Answer indices for each question i.e. [(question.start_index, question.end_index), ...] for all questions. Useful for slicing model predictions by question. See :ref:`schemas`.
    
    Returns:
        torch.Tensor: neg. log likelihood of shape (batch, question).
    """
    if careful:
        # some models give occasional nans for all predictions on a specific galaxy/row
        # these are inputs to the loss and only happen many epochs in so probably not a case of bad labels, but rather some instability during training
        # handle this by setting loss=0 for those rows and throwing a warning
        nan_prediction = torch.isnan(predictions) | torch.isinf(predictions)
        if nan_prediction.any():
            logging.warning(f'Found nan values in predictions: {predictions}')
        safety_value = torch.ones(1, device=predictions.device, dtype=predictions.dtype)  # fill with 1 everywhere i.e. fully uncertain
        predictions = torch.where(condition=nan_prediction, input=safety_value, other=predictions)

    # very important that question_index_groups is fixed and discrete, else tf.function autograph will mess up 
    q_losses = []
    # will give shape errors if model output dim is not labels dim, which can happen if losses.py substrings are missing an answer
    for q_n in range(len(question_index_groups)):
        q_indices = question_index_groups[q_n]
        q_start = q_indices[0]
        q_end = q_indices[1]
        q_loss = dirichlet_loss(labels[:, q_start:q_end+1], predictions[:, q_start:q_end+1])

        q_losses.append(q_loss)
    
    total_loss = torch.stack(q_losses, axis=1)

    return total_loss  # leave the reduction to pytorch lightning


def dirichlet_loss(labels_for_q, concentrations_for_q):
    """
    Negative log likelihood of ``labels_for_q`` being drawn from Dirichlet-Multinomial distribution with ``concentrations_for_q`` concentrations.
    This loss is for one question. Sum over multiple questions if needed (assumes independent).
    Applied by :meth:`calculate_multiquestion_loss`, above.

    Args:
        labels_for_q (tf.constant): observed labels (count of volunteer responses) of shape (batch, answer)
        concentrations_for_q (tf.constant): concentrations for multinomial-dirichlet predicting the observed labels, of shape (batch, answer)

    Returns:
        tf.constant: negative log. prob per galaxy, of shape (batch_dim).
    """
    # if you get a dimension mismatch here like [16, 2] vs. [64, 2], for example, check --shard-img-size is correct in train_model.py.
    # images may be being reshaped to the wrong expected size e.g. if they are saved as 32x32, but you put --shard-img-size=64,
    # you will get image batches of shape [N/4, 64, 64, 1] and hence have the wrong number of images vs. labels (and meaningless images)
    # so check --shard-img-size carefully!
    total_count = torch.sum(labels_for_q, axis=1)

    # pytorch dirichlet multinomial implementation will not accept zero total votes, need to handle separately
    return get_dirichlet_neg_log_prob(labels_for_q, total_count, concentrations_for_q)


    # # return tf.where(
    # #     tf.math.equal(total_count, 0),  # if total_count is 0 (i.e. no votes)
    # #     tf.constant(0.),  # prob is 1 and (neg) log prob is 0, with no gradients attached. broadcasts
    # #     get_dirichlet_neg_log_prob(labels_for_q, total_count, concentrations_for_q)  # else, calculate neg log prob
    # # )
    # # # slightly inefficient as get_dirichlet_neg_log_prob forward pass done for all, but avoids gradients
    # # # https://www.tensorflow.org/api_docs/python/tf/where
    # # works great, but about 50% slower than optimal

    # logging.info(total_count)

    # indices = torch.arange(0, len(total_count), dtype=torch.long)
    # indices_with_nonzero_counts = indices[total_count > 0]  # returns a tuple for some reason
    #     # torch.equal(total_count, torch.zeros(size=(1,)))

    # logging.info(indices_with_nonzero_counts)
    # # logging.info(type(indices_with_nonzero_counts))
    # # logging.info('Nonzero indices: {}'.format(indices_with_nonzero_counts.cpu().numpy()))
    
    # # may potentially need to deal with the situation where there are 0 valid indices?
    # nonzero_total_count = total_count[indices_with_nonzero_counts]
    # nonzero_labels_for_q = labels_for_q[indices_with_nonzero_counts]
    # nonzero_concentrations_for_q = concentrations_for_q[indices_with_nonzero_counts]

    # logging.info(nonzero_total_count)

    # neg_log_prob_of_indices_with_nonzero_counts = get_dirichlet_neg_log_prob(nonzero_labels_for_q, nonzero_total_count, nonzero_concentrations_for_q)
    # logging.info(neg_log_prob_of_indices_with_nonzero_counts)
    # # logging.info('Probs of nonzero indices: {}'.format(neg_log_prob_of_indices_with_nonzero_counts.numpy()))

    # # not needed, scatter_nd has 0 where no value is placed (and 0 as log prob = 0 as prob = 1)
    # # neg_log_prob_of_indices_with_zero_counts = tf.zeros_like(indices_with_zero_counts)

    # # now mix back together
    # # The backward pass is implemented only for src.shape == index.shape, true in my case (each neg lob prob goes exactly one place)
    # mixed_back = torch.zeros_like(total_count).scatter_(dim=0, index=indices_with_nonzero_counts, src=neg_log_prob_of_indices_with_nonzero_counts)
    # return mixed_back
    # # return tf.scatter_nd(indices_with_nonzero_counts, neg_log_prob_of_indices_with_nonzero_counts, shape=output_shape)


    # # return get_dirichlet_neg_log_prob(labels_for_q, total_count, concentrations_for_q)


def get_dirichlet_neg_log_prob(labels_for_q, total_count, concentrations_for_q):
    # https://docs.pyro.ai/en/stable/distributions.html#dirichletmultinomial
    # .int()s avoid rounding errors causing loss of around 1e-5 for questions with 0 votes
    dist = pyro.distributions.DirichletMultinomial(
        total_count=total_count.int(), concentration=concentrations_for_q, is_sparse=False, validate_args=True)
    return -dist.log_prob(labels_for_q.int())  # important minus sign



import warnings

import torch
import numpy as np
from typing import Optional


def cosine_schedule(
    step: int,
    max_steps: int,
    start_value: float,
    end_value: float,
    period: Optional[int] = None,
) -> float:
    """
    Use cosine decay to gradually modify start_value to reach target end_value during
    iterations.
    Copied from lightly library (thank you for open sourcing)

    Args:
        step:
            Current step number.
        max_steps:
            Total number of steps.
        start_value:
            Starting value.
        end_value:
            Target value.
        period (optional):
            The number of steps over which the cosine function completes a full cycle.
            If not provided, it defaults to max_steps.

    Returns:
        Cosine decay value.

    """
    if step < 0:
        raise ValueError("Current step number can't be negative")
    if max_steps < 1:
        raise ValueError("Total step number must be >= 1")
    if period is None and step > max_steps:
        warnings.warn(
            f"Current step number {step} exceeds max_steps {max_steps}.",
            category=RuntimeWarning,
        )
    if period is not None and period <= 0:
        raise ValueError("Period must be >= 1")

    decay: float
    if period is not None:  # "cycle" based on period, if provided
        decay = (
            end_value
            - (end_value - start_value) * (np.cos(2 * np.pi * step / period) + 1) / 2
        )
    elif max_steps == 1:
        # Avoid division by zero
        decay = end_value
    elif step == max_steps:
        # Special case for Pytorch Lightning which updates LR scheduler also for epoch
        # after last training epoch.
        decay = end_value
    else:
        decay = (
            end_value
            - (end_value - start_value)
            * (np.cos(np.pi * step / (max_steps - 1)) + 1)
            / 2
        )
    return decay


class CosineWarmupScheduler(torch.optim.lr_scheduler.LambdaLR):
    """Cosine warmup scheduler for learning rate.

    Args:
        optimizer:
            Optimizer object to schedule the learning rate.
        warmup_epochs:
            Number of warmup epochs or steps.
        max_epochs:
            Total number of training epochs or steps.
        last_epoch:
            The index of last epoch or step. Default: -1
        start_value:
            Starting learning rate scale. Default: 1.0
        end_value:
            Target learning rate scale. Default: 0.001
        verbose:
            If True, prints a message to stdout for each update. Default: False.

    Note: The `epoch` arguments do not necessarily have to be epochs. Any step or index
    can be used. The naming follows the Pytorch convention to use `epoch` for the steps
    in the scheduler.
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        warmup_epochs: int,
        max_epochs: int,
        last_epoch: int = -1,
        start_value: float = 1.0,
        end_value: float = 0.001,
        period: Optional[int] = None,
        verbose: bool = False,
    ) -> None:
        self.warmup_epochs = warmup_epochs
        self.max_epochs = max_epochs
        self.start_value = start_value
        self.end_value = end_value
        self.period = period
        super().__init__(
            optimizer=optimizer,
            lr_lambda=self.scale_lr,
            last_epoch=last_epoch,
            verbose=verbose,
        )

    def scale_lr(self, epoch: int) -> float:
        """
        Scale learning rate according to the current epoch number.

        Args:
            epoch:
                Current epoch number.

        Returns:
            Scaled learning rate.

        """
        if epoch < self.warmup_epochs:
            return self.start_value * (epoch + 1) / self.warmup_epochs
        elif self.period is not None:
            return cosine_schedule(
                step=epoch - self.warmup_epochs,
                max_steps=1,
                start_value=self.start_value,
                end_value=self.end_value,
                period=self.period,
            )
        else:
            return cosine_schedule(
                step=epoch - self.warmup_epochs,
                max_steps=self.max_epochs - self.warmup_epochs,
                start_value=self.start_value,
                end_value=self.end_value,
            )



# import numpy as np
# from scipy.special import gamma, loggamma, digamma, betaln  # log more numerically stable than gamma
# from scipy.stats import beta  # or could use tfp
# import tensorflow as tf
# import tensorflow_probability as tfp


# def entropy_upper_bound(arr_of_alphas, weights):
#     return entropy_given_components(arr_of_alphas, weights) - distance_term(arr_of_alphas, weights, distance_func=dirichlet_kl_div)


# def entropy_lower_bound(arr_of_alphas, weights):
#     return entropy_given_components(arr_of_alphas, weights) - distance_term(arr_of_alphas, weights, distance_func=lambda x, y: dirichlet_chernoff(x, y, lam=.5))


# def distance_term(answer_model, weights, distance_func):
#     n_dim = len(weights)
#     i_terms = np.zeros(len(weights))
#     for i in range(n_dim):
#         j_terms = np.zeros(len(weights))
#         for j in range(n_dim):
#             # the i=j term is not 0, don't skip it
#             distance = distance_func(answer_model[:, i], answer_model[:, j])
#             j_terms[j] = weights[j] * np.exp(-distance)
#         i_terms[i] = weights[i] * np.log(np.sum(j_terms))
#     return np.sum(i_terms)


# # def mbeta(x):
# #     result = np.prod(gamma(x))/gamma(x.sum())
# #     if np.isnan(result):
# #         raise ValueError(x)
# #     else:
# #         return result

# # def dirichlet_chernoff(a, b, lam):
# #     exp_neg_chern = mbeta(lam*a + (1-lam)*b) / (mbeta(a) ** lam * mbeta(b) ** (1-lam))
# #     chern = -np.log(exp_neg_chern)
# #     return chern

# def mbeta(x):
#     result = np.prod(gamma(x))/gamma(x.sum())
#     if np.isnan(result):
#         raise ValueError(x)
#     else:
#         return result

# def log_mbeta(x):
#     return np.sum(loggamma(x)) - loggamma(np.sum(x))  # nice symmetry

# # def dirichlet_chernoff(a, b, lam):
# #     exp_neg_chern = mbeta(lam*a + (1-lam)*b) / (mbeta(a) ** lam * mbeta(b) ** (1-lam))
# #     chern = -np.log(exp_neg_chern)
# #     return chern

# def dirichlet_chernoff(a, b, lam):
#     neg_chern = log_mbeta(lam * a + (1-lam) * b) - lam*log_mbeta(a) - (1-lam)*log_mbeta(b)
#     return -neg_chern

# # copied from tfp, probably not needed
# # @tf.function
# def dirichlet_kl_div(concentration1, concentration2):
#     # concentration1 = tf.convert_to_tensor(concentration1)
#     # concentration2 = tf.convert_to_tensor(concentration2)
#     digamma_sum_d1 = tf.math.digamma(
#         tf.reduce_sum(concentration1, axis=-1, keepdims=True))
#     digamma_diff = tf.math.digamma(concentration1) - digamma_sum_d1
#     concentration_diff = concentration1 - concentration2

#     return (
#         tf.reduce_sum(concentration_diff * digamma_diff, axis=-1) -
#         tf.math.lbeta(concentration1) + tf.math.lbeta(concentration2))

# # pure python version
# # def dirichlet_kl_div(concentration1, concentration2):
# #     digamma_sum_d1 = digamma(concentration1.sum(axis=-1))
# #     digamma_diff = digamma(concentration1) - digamma_sum_d1
# #     concentration_diff = concentration1 - concentration2
# #     positive_comp = (concentration_diff * digamma_diff).sum(axis=-1)
# #     # negative_comp = betaln(*concentration1) + betaln(*concentration2)
# #     negative_comp = tf.math.lbeta(concentration1) + tf.math.lbeta(concentration2)
# #     return positive_comp - negative_comp.numpy()


# def entropy_given_components(answer_model, weights):
#     list_of_alphas = tf.transpose(answer_model)
#     return np.sum(weights * np.array([tfp.distributions.Dirichlet(alpha).entropy() for alpha in list_of_alphas]))




# mostly refactored out into galaxy-datasets.shared.label_metadata.
# zoobot depends on galaxy-datasets, but galaxy-datasets does not depend on zoobot

from galaxy_datasets.shared.label_metadata import extract_questions_and_label_cols



"""
Non-orthogonal schemas (i.e. without the -dr5, -gz2, etc.)
Not recommended for use with ML due to potential overlap between questions from different surveys.
May still be convenient for exploring catalogs (just to save typing -dr5 etc all the time)
"""

decals_pairs = {
    'smooth-or-featured': ['_smooth', '_featured-or-disk', '_artifact'],
    'disk-edge-on': ['_yes', '_no'],
    'has-spiral-arms': ['_yes', '_no'],
    'bar': ['_strong', '_weak', '_no'],
    'bulge-size': ['_dominant', '_large', '_moderate', '_small', '_none'],
    'how-rounded': ['_round', '_in-between', '_cigar-shaped'],
    'edge-on-bulge': ['_boxy', '_none', '_rounded'],
    'spiral-winding': ['_tight', '_medium', '_loose'],
    'spiral-arm-count': ['_1', '_2', '_3', '_4', '_more-than-4', '_cant-tell'],
    'merging': ['_none', '_minor-disturbance', '_major-disturbance', '_merger']
}
decals_questions, decals_label_cols = extract_questions_and_label_cols(decals_pairs)


gz2_pairs = {
    'smooth-or-featured': ['_smooth', '_featured-or-disk', '_artifact'],
    'disk-edge-on': ['_yes', '_no'],
    'has-spiral-arms': ['_yes', '_no'],
    'bar': ['_yes', '_no'],
    'bulge-size': ['_dominant', '_obvious', '_just-noticeable', '_no'],
    'something-odd': ['_yes', '_no'],
    'how-rounded': ['_round', '_in-between', '_cigar'],
    'bulge-shape': ['_round', '_boxy', '_no-bulge'],
    'spiral-winding': ['_tight', '_medium', '_loose'],
    'spiral-arm-count': ['_1', '_2', '_3', '_4', '_more-than-4', '_cant-tell']
}
gz2_questions, gz2_label_cols = extract_questions_and_label_cols(gz2_pairs)


# useful for development/debugging
gz2_partial_ortho_pairs = {
    'smooth-or-featured-gz2': ['_smooth', '_featured-or-disk']
}
gz2_partial_ortho_questions, gz2_partial_ortho_label_cols = extract_questions_and_label_cols(gz2_partial_ortho_pairs)




import logging
import os

import tensorflow as tf
from tensorflow.keras import layers

from zoobot.tensorflow.training import training_config


def run_finetuning(config, encoder, train_dataset, val_dataset, test_dataset, save_dir):

    new_head = linear_classifier(config['finetune']['encoder_dim'], config['finetune']['label_dim'])

    img_size = config['finetune']['img_size']

    """
    Retrain the model. Only the new head will train as the rest is frozen.
    """


    encoder.trainable = False

    # stick the new head on the pretrained base model
    model = tf.keras.Sequential([
      tf.keras.layers.InputLayer(input_shape=(img_size, img_size, 1)),
      encoder,
      new_head
    ])

    loss = tf.keras.losses.CategoricalCrossentropy(from_logits=True)

    model.compile(
        loss=loss,
        optimizer=tf.keras.optimizers.Adam(learning_rate=0.001),  # normal learning rate is okay
        metrics=['accuracy']
    )
    model.summary()

    trainer = training_config.Trainer(
        # parameters for how to train e.g. epochs, patience
        log_dir=os.path.join(save_dir, 'head_only'),
        epochs=config['finetune']['n_epochs'],
        patience=config['finetune']['patience']
    )

    model_with_trained_head = trainer.fit(
        model,
        train_dataset,
        val_dataset,
        eager=False
    )

    logging.info('Head finetuning complete')

    if config['finetune']['n_layers'] == 0:
        logging.info('n_layers = 0: not finetuning lower layers')
        return model_with_trained_head  

    logging.info('Unfreezing layers')
    # you can unfreeze layers like so:
    unfreeze_model(model, unfreeze_names=['top'])
    # or more...
    # utils.unfreeze_model(model, unfreeze_names=['top', 'block7'])
    # utils.unfreeze_model(model, unfreeze_names=['top', 'block7', 'block6'])
    # utils.unfreeze_model(model, unfreeze_names=['top', 'block7', 'block6', 'block5'])
    # utils.unfreeze_model(model, unfreeze_names=['top', 'block7', 'block6', 'block5', 'block4'])
    # utils.unfreeze_model(model, unfreeze_names=[], unfreeze_all=True)


    logging.info('Recompiling with lower learning rate and trainable upper layers')
    model_with_trained_head.compile(
        loss=tf.keras.losses.binary_crossentropy,
        optimizer=tf.keras.optimizers.Adam(learning_rate=1e-5),  # 10x lower initial learning rate (adam will adapt anyway)
        metrics=['accuracy']
    )

    model_with_trained_head.summary(print_fn=logging.info)

    trainer = training_config.Trainer(
        # parameters for how to train e.g. epochs, patience
        log_dir=os.path.join(save_dir, 'full'),
        epochs=config['finetune']['n_epochs'],
        patience=config['finetune']['patience']
    )

    model_with_trained_lower_layers = trainer.fit(
        model_with_trained_head,
        train_dataset,
        val_dataset,
        eager=False
    )
    
    logging.info('Finetuning complete')
    return model_with_trained_lower_layers
    

    
    """
    Well done!
    
    You can now use your finetuned models to make predictions on new data..
    See make_predictions.py for a self-contained example.
    """


def linear_classifier(input_dim, output_dim):
    return tf.keras.Sequential([
      # TODO move pooling
      tf.keras.layers.InputLayer(input_shape=(input_dim)),  # base model dim after GlobalAveragePooling (ignoring batch)
      tf.keras.layers.Dense(output_dim, name='logits')  # output should be N neurons w/ softmax for N-class classification
      # layers.Dense(3, activation="softmax", name="softmax_output")  # ...or 
    ])



def freeze_model(model):
    # Freeze the pretrained weights
    # inplace
    model.trainable = False


def unfreeze_model(model, unfreeze_names=['block7', 'top'], unfreeze_all=False):
    if unfreeze_all and (len(unfreeze_names) > 0):
        logging.warning('unfreeze_all is True; ignoring unfreeze_names and unfreezing all layers')
    # https://keras.io/examples/vision/image_classification_efficientnet_fine_tuning/
    # required for any layer to be trainable.
    # however, setting to True sets *every* layer trainable (why, tf, why...)
    # so need to then set each layer individually trainable or not trainable below
    model.trainable = True  # everything trainable, recursively.

    for layer in model.layers:
        # recursive
        # if isinstance(layer, tf.keras.Sequential) or isinstance(layer, tf.python.keras.engine.functional.Functional):  # layer is itself a model (effnet is functional due to residual connections)
        if isinstance(layer, tf.keras.Model):  # includes subclasses Sequential and Functional
            unfreeze_model(layer, unfreeze_names=unfreeze_names, unfreeze_all=unfreeze_all)  # recursive

        elif any([layer.name.startswith(name) for name in unfreeze_names]) or unfreeze_all:
            if isinstance(layer, layers.BatchNormalization):
                logging.debug('freezing batch norm layer {}'.format(layer.name))
                layer.trainable = False
            else:
                logging.debug('unfreezing {}'.format(layer.name))
                layer.trainable = True
                # print('Freezing batch norm layer')
                # https://www.tensorflow.org/api_docs/python/tf/keras/layers/BatchNormalization?version=stable#note_that_2
                # this will also switch layer to inference mode from tf2, no need to separately pass training=False
        else:
            logging.warning('Layer {} ({}) not in unfreeze list - freezing by default'.format(layer.name, layer))
            layer.trainable = False  # not a recursive call, and not with a name to unfreeze

    # model will be trainable next time it is compiled

def check_batchnorm_frozen(model):
    for layer in model.layers:
        print(layer)
        if isinstance(layer, tf.keras.Model):
            check_batchnorm_frozen(layer)
        elif isinstance(layer, layers.BatchNormalization):
            assert not layer.trainable
            print('checks out')



    # import numpy as np
    # import pandas as pd


    # paths_pred = paths_val  # TODO for simplicitly I'll just make more predictions on the validation images, but you'll want to change this
    # raw_pred_dataset = image_datasets.get_image_dataset(paths_pred, file_format=file_format, requested_img_size=requested_img_size, batch_size=batch_size)

    # ordered_paths = [x.numpy().decode('utf8') for batch in raw_pred_dataset for x in batch['id_str']]

    # # must exactly match the preprocessing you used for training
    # pred_config = preprocess.PreprocessingConfig(
    #   label_cols=[],  # image_datasets.get_image_dataset will put the labels arg under 'label' key for each batch
    #   input_size=requested_img_size,
    #   make_greyscale=True,
    #   # normalise_from_uint8=True,
    #   permute_channels=False
    # )
    # pred_dataset = preprocess.preprocess_dataset(raw_pred_dataset, pred_config)

    # predictions = model.predict(pred_dataset)

    # data = [{'prediction': float(prediction), 'image_loc': local_png_loc} for prediction, local_png_loc in zip(predictions, ordered_paths)]
    # pred_df = pd.DataFrame(data=data)
  
    # example_predictions_loc = 'results/finetune_minimal/example_predictions.csv'
    # pred_df.to_csv(example_predictions_loc, index=False)
    # logging.info(f'Example predictions saved to {example_predictions_loc}')


import pytest

import timm
import torch


def test_get_encoder():
    model = timm.create_model("hf_hub:mwalmsley/zoobot-encoder-efficientnet_b0", pretrained=True)
    assert model(torch.rand(1, 3, 224, 224)).shape == (1, 1280)


def test_get_finetuned():
    # checkpoint_loc = 'https://huggingface.co/mwalmsley/zoobot-finetuned-is_tidal/resolve/main/3.ckpt' pickle problem via lightning
    # checkpoint_loc = '/home/walml/Downloads/3.ckpt'  # works when downloaded manually

    from huggingface_hub import hf_hub_download

    REPO_ID = "mwalmsley/zoobot-finetuned-is_tidal"
    FILENAME = "FinetuneableZoobotClassifier.ckpt"

    downloaded_loc = hf_hub_download(
        repo_id=REPO_ID,
        filename=FILENAME,
    )
    from zoobot.pytorch.training import finetune
    model = finetune.FinetuneableZoobotClassifier.load_from_checkpoint(downloaded_loc, map_location='cpu') #  hub_name='hf_hub:mwalmsley/zoobot-encoder-convnext_nano',
    assert model(torch.rand(1, 3, 224, 224)).shape == (1, 2)

def test_get_finetuned_class_method():

    from zoobot.pytorch.training import finetune

    model = finetune.FinetuneableZoobotClassifier.load_from_name('mwalmsley/zoobot-finetuned-is_tidal', map_location='cpu')
    assert model(torch.rand(1, 3, 224, 224)).shape == (1, 2)

# def test_get_finetuned_from_local():
#     # checkpoint_loc = '/home/walml/repos/zoobot/tests/convnext_nano_finetuned_linear_is-lsb.ckpt'
#     checkpoint_loc = '/home/walml/repos/zoobot-foundation/results/finetune/is-lsb/debug/checkpoints/4.ckpt'
    
#     from zoobot.pytorch.training import finetune
#     # if originally trained with a direct in-memory checkpoint, must specify the hub name manually. otherwise it's saved as an hparam.
#     model = finetune.FinetuneableZoobotClassifier.load_from_checkpoint(checkpoint_loc, map_location='cpu') # hub_name='hf_hub:mwalmsley/zoobot-encoder-convnext_nano', )
#     assert model(torch.rand(1, 3, 224, 224)).shape == (1, 2)



import logging

import tensorflow as tf

from zoobot.tensorflow.estimators import efficientnet_standard


def define_headless_efficientnet(input_shape=None, get_effnet=efficientnet_standard.EfficientNetB0, use_imagenet_weights=False, **kwargs):
    """
    Define efficientnet model to train.
    Thin wrapper around ``get_effnet``, an efficientnet creation function from ``efficientnet_standard``, that ensures the appropriate args.

    Additional keyword arguments are passed to ``get_effnet``.

    Args:
        input_shape (tuple, optional): Expected input shape e.g. (224, 224, 1). Defaults to None.
        get_effnet (function, optional): Efficientnet creation function from ``efficientnet_standard``. Defaults to efficientnet_standard.EfficientNetB0.
    
    Returns:
        [type]: [description]
    """
    model = tf.keras.models.Sequential(name='headless_efficientnet')
    logging.info('Building efficientnet to expect input {}, after any preprocessing layers'.format(input_shape))


    if use_imagenet_weights:
        logging.warning('Using imagenet weights - not recommended!')
        weights = 'imagenet'  # split variable names to be clear this isn't one of my checkpoints to load
    else:
        weights = None

    # classes probably does nothing without include_top
    effnet = get_effnet(
        input_shape=input_shape,
        weights=weights,
        include_top=False,  # no final three layers: pooling, dropout and dense
        classes=None,  # headless so has no effect
        **kwargs
    )
    model.add(effnet)

    return model


def custom_top_dirichlet(output_dim):
    """
    Final dense layer used in GZ DECaLS (after global pooling). 
    ``output_dim`` neurons with an activation of ``tf.nn.sigmoid(x) * 100. + 1.``, chosen to ensure 1-100 output range
    This range is suitable for parameters of Dirichlet distribution.

    Use with Functional API e.g. x = custom_top_dirichlet(output_dim)(x)

    Args:
        output_dim (int): Dimension of dense layer e.g. 34 for decision tree with 34 answers

    Returns:
        (tf.keras.layers.Dense): suitable for predicting Dirichlet distributions, as above.
    """
    return tf.keras.layers.Dense(output_dim, activation=lambda x: tf.nn.sigmoid(x) * 100. + 1.)  # one params per answer, 1-100 range
